#!/usr/bin/python

# Copyright (c) 2019, Guillaume Martinez (lunik@tiwabbit.fr)
# Copyright (c) 2018, Marcus Watkins <marwatk@marcuswatkins.net>
# Based on code:
# Copyright (c) 2013, Phillip Gentry <phillip@cx.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_hook
short_description: Manages GitLab project hooks
description:
  - Adds, updates and removes project hook.
author:
  - Marcus Watkins (@marwatk)
  - Guillaume Martinez (@Lunik)
requirements:
  - python-gitlab python module
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  project:
    description:
      - ID or Full path of the project in the form of group/name.
    required: true
    type: str
  hook_url:
    description:
      - The URL that you want GitLab to post to, this is used as the primary key for updates and deletion.
    required: true
    type: str
  state:
    description:
      - When V(present) the hook is updated to match the input or created if it does not exist.
      - When V(absent) hook is deleted if it exists.
    default: present
    type: str
    choices: ["present", "absent"]
  push_events:
    description:
      - Trigger hook on push events.
    type: bool
    default: true
  push_events_branch_filter:
    description:
      - Branch name of wildcard to trigger hook on push events.
    type: str
    version_added: '0.2.0'
    default: ''
  issues_events:
    description:
      - Trigger hook on issues events.
    type: bool
    default: false
  merge_requests_events:
    description:
      - Trigger hook on merge requests events.
    type: bool
    default: false
  tag_push_events:
    description:
      - Trigger hook on tag push events.
    type: bool
    default: false
  note_events:
    description:
      - Trigger hook on note events or when someone adds a comment.
    type: bool
    default: false
  job_events:
    description:
      - Trigger hook on job events.
    type: bool
    default: false
  pipeline_events:
    description:
      - Trigger hook on pipeline events.
    type: bool
    default: false
  wiki_page_events:
    description:
      - Trigger hook on wiki events.
    type: bool
    default: false
  releases_events:
    description:
      - Trigger hook on release events.
    type: bool
    version_added: '8.4.0'
  hook_validate_certs:
    description:
      - Whether GitLab performs SSL verification when triggering the hook.
    type: bool
    default: false
    aliases: [enable_ssl_verification]
  token:
    description:
      - Secret token to validate hook messages at the receiver.
      - If this is present it always results in a change as it cannot be retrieved from GitLab.
      - It shows up in the C(X-GitLab-Token) HTTP request header.
    type: str
"""

EXAMPLES = r"""
- name: "Adding a project hook"
  community.general.gitlab_hook:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    project: "my_group/my_project"
    hook_url: "https://my-ci-server.example.com/gitlab-hook"
    state: present
    push_events: true
    tag_push_events: true
    token: "my-super-secret-token-that-my-ci-server-will-check"

- name: "Delete the previous hook"
  community.general.gitlab_hook:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    project: "my_group/my_project"
    hook_url: "https://my-ci-server.example.com/gitlab-hook"
    state: absent

- name: "Delete a hook by numeric project id"
  community.general.gitlab_hook:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    project: 10
    hook_url: "https://my-ci-server.example.com/gitlab-hook"
    state: absent
"""

RETURN = r"""
msg:
  description: Success or failure message.
  returned: always
  type: str
  sample: "Success"

result:
  description: JSON parsed response from the server.
  returned: always
  type: dict

error:
  description: The error message returned by the GitLab API.
  returned: failed
  type: str
  sample: "400: path is already in use"

hook:
  description: API object.
  returned: always
  type: dict
"""

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    find_project,
    gitlab_authentication,
    list_all_kwargs,
)


class GitLabHook:
    def __init__(self, module, gitlab_instance):
        self._module = module
        self._gitlab = gitlab_instance
        self.hook_object = None

    """
    @param project Project Object
    @param hook_url Url to call on event
    @param description Description of the group
    @param parent Parent group full path
    """

    def create_or_update_hook(self, project, hook_url, options):
        changed = False

        hook_arguments = {
            "url": hook_url,
            "push_events": options["push_events"],
            "push_events_branch_filter": options["push_events_branch_filter"],
            "issues_events": options["issues_events"],
            "merge_requests_events": options["merge_requests_events"],
            "tag_push_events": options["tag_push_events"],
            "note_events": options["note_events"],
            "job_events": options["job_events"],
            "pipeline_events": options["pipeline_events"],
            "wiki_page_events": options["wiki_page_events"],
            "enable_ssl_verification": options["enable_ssl_verification"],
            "token": options["token"],
        }

        # Because we have already call userExists in main()
        if self.hook_object is None:
            if options["releases_events"] is not None:
                hook_arguments["releases_events"] = options["releases_events"]
            hook = self.create_hook(project, hook_arguments)
            changed = True
        else:
            update_arguments = hook_arguments.copy()
            update_arguments["releases_events"] = options["releases_events"]
            changed, hook = self.update_hook(self.hook_object, update_arguments)

        self.hook_object = hook
        if changed:
            if self._module.check_mode:
                self._module.exit_json(changed=True, msg=f"Successfully created or updated the hook {hook_url}")

            try:
                hook.save()
            except Exception as e:
                self._module.fail_json(msg=f"Failed to update hook: {e} ")

        return changed

    """
    @param project Project Object
    @param arguments Attributes of the hook
    """

    def create_hook(self, project, arguments):
        if self._module.check_mode:
            return True

        hook = project.hooks.create(arguments)

        return hook

    """
    @param hook Hook Object
    @param arguments Attributes of the hook
    """

    def update_hook(self, hook, arguments):
        changed = False

        for arg_key, arg_value in arguments.items():
            if arg_value is not None:
                if getattr(hook, arg_key, None) != arg_value:
                    setattr(hook, arg_key, arg_value)
                    changed = True

        return (changed, hook)

    """
    @param project Project object
    @param hook_url Url to call on event
    """

    def find_hook(self, project, hook_url):
        for hook in project.hooks.list(**list_all_kwargs):
            if hook.url == hook_url:
                return hook

    """
    @param project Project object
    @param hook_url Url to call on event
    """

    def exists_hook(self, project, hook_url):
        # When project exists, object will be stored in self.project_object.
        hook = self.find_hook(project, hook_url)
        if hook:
            self.hook_object = hook
            return True
        return False

    def delete_hook(self):
        if not self._module.check_mode:
            self.hook_object.delete()


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            state=dict(type="str", default="present", choices=["absent", "present"]),
            project=dict(type="str", required=True),
            hook_url=dict(type="str", required=True),
            push_events=dict(type="bool", default=True),
            push_events_branch_filter=dict(type="str", default=""),
            issues_events=dict(type="bool", default=False),
            merge_requests_events=dict(type="bool", default=False),
            tag_push_events=dict(type="bool", default=False),
            note_events=dict(type="bool", default=False),
            job_events=dict(type="bool", default=False),
            pipeline_events=dict(type="bool", default=False),
            wiki_page_events=dict(type="bool", default=False),
            releases_events=dict(type="bool"),
            hook_validate_certs=dict(type="bool", default=False, aliases=["enable_ssl_verification"]),
            token=dict(type="str", no_log=True),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
        ],
        required_together=[["api_username", "api_password"]],
        required_one_of=[["api_username", "api_token", "api_oauth_token", "api_job_token"]],
        supports_check_mode=True,
    )

    # check prerequisites and connect to gitlab server
    gitlab_instance = gitlab_authentication(module)

    state = module.params["state"]
    project_identifier = module.params["project"]
    hook_url = module.params["hook_url"]
    push_events = module.params["push_events"]
    push_events_branch_filter = module.params["push_events_branch_filter"]
    issues_events = module.params["issues_events"]
    merge_requests_events = module.params["merge_requests_events"]
    tag_push_events = module.params["tag_push_events"]
    note_events = module.params["note_events"]
    job_events = module.params["job_events"]
    pipeline_events = module.params["pipeline_events"]
    wiki_page_events = module.params["wiki_page_events"]
    releases_events = module.params["releases_events"]
    enable_ssl_verification = module.params["hook_validate_certs"]
    hook_token = module.params["token"]

    gitlab_hook = GitLabHook(module, gitlab_instance)

    project = find_project(gitlab_instance, project_identifier)

    if project is None:
        module.fail_json(msg=f"Failed to create hook: project {project_identifier} doesn't exists")

    hook_exists = gitlab_hook.exists_hook(project, hook_url)

    if state == "absent":
        if hook_exists:
            gitlab_hook.delete_hook()
            module.exit_json(changed=True, msg=f"Successfully deleted hook {hook_url}")
        else:
            module.exit_json(changed=False, msg="Hook deleted or does not exists")

    if state == "present":
        if gitlab_hook.create_or_update_hook(
            project,
            hook_url,
            {
                "push_events": push_events,
                "push_events_branch_filter": push_events_branch_filter,
                "issues_events": issues_events,
                "merge_requests_events": merge_requests_events,
                "tag_push_events": tag_push_events,
                "note_events": note_events,
                "job_events": job_events,
                "pipeline_events": pipeline_events,
                "wiki_page_events": wiki_page_events,
                "releases_events": releases_events,
                "enable_ssl_verification": enable_ssl_verification,
                "token": hook_token,
            },
        ):
            module.exit_json(
                changed=True,
                msg=f"Successfully created or updated the hook {hook_url}",
                hook=gitlab_hook.hook_object._attrs,
            )
        else:
            module.exit_json(
                changed=False, msg=f"No need to update the hook {hook_url}", hook=gitlab_hook.hook_object._attrs
            )


if __name__ == "__main__":
    main()



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: linode
author:
  - Luke Murphy (@decentral1se)
short_description: Ansible dynamic inventory plugin for Linode
requirements:
  - linode_api4 >= 2.0.0
description:
  - Reads inventories from the Linode API v4.
  - Uses a YAML configuration file that ends with linode.(yml|yaml).
  - Linode labels are used by default as the hostnames.
  - The default inventory groups are built from groups (deprecated by Linode) and not tags.
extends_documentation_fragment:
  - ansible.builtin.constructed
  - ansible.builtin.inventory_cache
options:
  cache:
    version_added: 4.5.0
  cache_plugin:
    version_added: 4.5.0
  cache_timeout:
    version_added: 4.5.0
  cache_connection:
    version_added: 4.5.0
  cache_prefix:
    version_added: 4.5.0
  plugin:
    description: Marks this as an instance of the 'linode' plugin.
    type: string
    required: true
    choices: ['linode', 'community.general.linode']
  ip_style:
    description: Populate hostvars with all information available from the Linode APIv4.
    type: string
    default: plain
    choices:
      - plain
      - api
    version_added: 3.6.0
  access_token:
    description: The Linode account personal access token.
    type: string
    required: true
    env:
      - name: LINODE_ACCESS_TOKEN
  regions:
    description: Populate inventory with instances in this region.
    default: []
    type: list
    elements: string
  tags:
    description: Populate inventory only with instances which have at least one of the tags listed here.
    default: []
    type: list
    elements: string
    version_added: 2.0.0
  types:
    description: Populate inventory with instances with this type.
    default: []
    type: list
    elements: string
  strict:
    version_added: 2.0.0
  compose:
    version_added: 2.0.0
  groups:
    version_added: 2.0.0
  keyed_groups:
    version_added: 2.0.0
"""

EXAMPLES = r"""
---
# Minimal example. `LINODE_ACCESS_TOKEN` is exposed in environment.
plugin: community.general.linode

---
# You can use Jinja to template the access token.
plugin: community.general.linode
access_token: "{{ lookup('ini', 'token', section='your_username', file='~/.config/linode-cli') }}"
# For older Ansible versions, you need to write this as:
# access_token: "{{ lookup('ini', 'token section=your_username file=~/.config/linode-cli') }}"

---
# Example with regions, types, groups and access token
plugin: community.general.linode
access_token: foobar
regions:
  - eu-west
types:
  - g5-standard-2

---
# Example with keyed_groups, groups, and compose
plugin: community.general.linode
access_token: foobar
keyed_groups:
  - key: tags
    separator: ''
  - key: region
    prefix: region
groups:
  webservers: "'web' in (tags|list)"
  mailservers: "'mail' in (tags|list)"
compose:
  # By default, Ansible tries to connect to the label of the instance.
  # Since that might not be a valid name to connect to, you can
  # replace it with the first IPv4 address of the linode as follows:
  ansible_ssh_host: ipv4[0]
  ansible_port: 2222

---
# Example where control traffic limited to internal network
plugin: community.general.linode
access_token: foobar
ip_style: api
compose:
  ansible_host: "ipv4 | community.general.json_query('[?public==`false`].address') | first"
"""

from ansible.errors import AnsibleError
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, Constructable

from ansible_collections.community.general.plugins.plugin_utils._unsafe import make_unsafe

try:
    from linode_api4 import LinodeClient
    from linode_api4.errors import ApiError as LinodeApiError
    from linode_api4.objects.linode import Instance

    HAS_LINODE = True
except ImportError:
    HAS_LINODE = False


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):
    NAME = "community.general.linode"

    def _build_client(self, loader):
        """Build the Linode client."""

        access_token = self.get_option("access_token")
        if self.templar.is_template(access_token):
            access_token = self.templar.template(variable=access_token)

        if access_token is None:
            raise AnsibleError("Could not retrieve Linode access token from plugin configuration sources")

        self.client = LinodeClient(access_token)

    def _get_instances_inventory(self):
        """Retrieve Linode instance information from cloud inventory."""
        try:
            self.instances = self.client.linode.instances()
        except LinodeApiError as exception:
            raise AnsibleError(f"Linode client raised: {exception}") from exception

    def _add_groups(self):
        """Add Linode instance groups to the dynamic inventory."""
        self.linode_groups = {instance.group for instance in self.instances if instance.group}

        for linode_group in self.linode_groups:
            self.inventory.add_group(linode_group)

    def _filter_by_config(self):
        """Filter instances by user specified configuration."""
        regions = self.get_option("regions")
        if regions:
            self.instances = [instance for instance in self.instances if instance.region.id in regions]

        types = self.get_option("types")
        if types:
            self.instances = [instance for instance in self.instances if instance.type.id in types]

        tags = self.get_option("tags")
        if tags:
            self.instances = [instance for instance in self.instances if any(tag in instance.tags for tag in tags)]

    def _add_instances_to_groups(self):
        """Add instance names to their dynamic inventory groups."""
        for instance in self.instances:
            self.inventory.add_host(make_unsafe(instance.label), group=instance.group)

    def _add_hostvars_for_instances(self):
        """Add hostvars for instances in the dynamic inventory."""
        ip_style = self.get_option("ip_style")
        for instance in self.instances:
            hostvars = instance._raw_json
            hostname = make_unsafe(instance.label)
            for hostvar_key in hostvars:
                if ip_style == "api" and hostvar_key in ["ipv4", "ipv6"]:
                    continue
                self.inventory.set_variable(hostname, hostvar_key, make_unsafe(hostvars[hostvar_key]))
            if ip_style == "api":
                ips = instance.ips.ipv4.public + instance.ips.ipv4.private
                ips += [instance.ips.ipv6.slaac, instance.ips.ipv6.link_local]
                ips += instance.ips.ipv6.pools

                for ip_type in {ip.type for ip in ips}:
                    self.inventory.set_variable(
                        hostname, ip_type, make_unsafe(self._ip_data([ip for ip in ips if ip.type == ip_type]))
                    )

    def _ip_data(self, ip_list):
        data = []
        for ip in list(ip_list):
            data.append(
                {
                    "address": ip.address,
                    "subnet_mask": ip.subnet_mask,
                    "gateway": ip.gateway,
                    "public": ip.public,
                    "prefix": ip.prefix,
                    "rdns": ip.rdns,
                    "type": ip.type,
                }
            )
        return data

    def _cacheable_inventory(self):
        return [i._raw_json for i in self.instances]

    def populate(self):
        strict = self.get_option("strict")

        self._filter_by_config()

        self._add_groups()
        self._add_instances_to_groups()
        self._add_hostvars_for_instances()
        for instance in self.instances:
            hostname = make_unsafe(instance.label)
            variables = self.inventory.get_host(hostname).get_vars()
            self._add_host_to_composed_groups(self.get_option("groups"), variables, hostname, strict=strict)
            self._add_host_to_keyed_groups(self.get_option("keyed_groups"), variables, hostname, strict=strict)
            self._set_composite_vars(self.get_option("compose"), variables, hostname, strict=strict)

    def verify_file(self, path):
        """Verify the Linode configuration file.

        Return true/false if the config-file is valid for this plugin

        Args:
            str(path): path to the config
        Kwargs:
            None
        Raises:
            None
        Returns:
            bool(valid): is valid config file"""
        valid = False
        if super().verify_file(path):
            if path.endswith(("linode.yaml", "linode.yml")):
                valid = True
            else:
                self.display.vvv('Inventory source not ending in "linode.yaml" or "linode.yml"')
        return valid

    def parse(self, inventory, loader, path, cache=True):
        """Dynamically parse Linode the cloud inventory."""
        super().parse(inventory, loader, path)
        self.instances = None

        if not HAS_LINODE:
            raise AnsibleError("the Linode dynamic inventory plugin requires linode_api4.")

        self._read_config_data(path)

        cache_key = self.get_cache_key(path)

        if cache:
            cache = self.get_option("cache")

        update_cache = False
        if cache:
            try:
                self.instances = [Instance(None, i["id"], i) for i in self._cache[cache_key]]
            except KeyError:
                update_cache = True

        # Check for None rather than False in order to allow
        # for empty sets of cached instances
        if self.instances is None:
            self._build_client(loader)
            self._get_instances_inventory()

        if update_cache:
            self._cache[cache_key] = self._cacheable_inventory()

        self.populate()



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the nxos_ospf_interfaces module
"""


class Ospf_interfacesArgs(object):  # pylint: disable=R0903
    """The arg spec for the nxos_ospf_interfaces module"""

    argument_spec = {
        "running_config": {"type": "str"},
        "config": {
            "type": "list",
            "elements": "dict",
            "options": {
                "name": {"type": "str", "required": True},
                "address_family": {
                    "type": "list",
                    "elements": "dict",
                    "mutually_exclusive": [["passive_interface", "default_passive_interface"]],
                    "options": {
                        "afi": {
                            "type": "str",
                            "choices": ["ipv4", "ipv6"],
                            "required": True,
                        },
                        "processes": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "process_id": {
                                    "type": "str",
                                    "required": True,
                                },
                                "area": {
                                    "type": "dict",
                                    "options": {
                                        "area_id": {
                                            "type": "str",
                                            "required": True,
                                        },
                                        "secondaries": {"type": "bool"},
                                    },
                                },
                                "multi_areas": {
                                    "type": "list",
                                    "elements": "str",
                                },
                            },
                        },
                        "multi_areas": {"type": "list", "elements": "str"},
                        "authentication": {
                            "type": "dict",
                            "options": {
                                "key_chain": {"type": "str", "no_log": False},
                                "message_digest": {"type": "bool"},
                                "enable": {"type": "bool"},
                                "null_auth": {"type": "bool"},
                            },
                        },
                        "authentication_key": {
                            "type": "dict",
                            "no_log": False,
                            "options": {
                                "encryption": {"type": "int"},
                                "key": {
                                    "type": "str",
                                    "required": True,
                                    "no_log": True,
                                },
                            },
                        },
                        "message_digest_key": {
                            "type": "dict",
                            "no_log": False,
                            "options": {
                                "key_id": {"type": "int", "required": True},
                                "encryption": {"type": "int"},
                                "key": {
                                    "type": "str",
                                    "required": True,
                                    "no_log": True,
                                },
                            },
                        },
                        "cost": {"type": "int"},
                        "dead_interval": {"type": "int"},
                        "hello_interval": {"type": "int"},
                        "instance": {"type": "int"},
                        "mtu_ignore": {"type": "bool"},
                        "network": {
                            "type": "str",
                            "choices": ["broadcast", "point-to-point"],
                        },
                        "default_passive_interface": {"type": "bool"},
                        "passive_interface": {"type": "bool"},
                        "priority": {"type": "int"},
                        "retransmit_interval": {"type": "int"},
                        "shutdown": {"type": "bool"},
                        "transmit_delay": {"type": "int"},
                    },
                },
            },
        },
        "state": {
            "type": "str",
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "gathered",
                "parsed",
                "rendered",
            ],
            "default": "merged",
        },
    }  # pylint: disable=C0301



# -*- coding: utf-8 -*-
# (c) 2018 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


DOCUMENTATION = '''
    name: cgroup_perf_recap
    type: aggregate
    requirements:
      - whitelist in configuration
      - cgroups
    short_description: Profiles system activity of tasks and full execution using cgroups
    description:
        - This is an ansible callback plugin utilizes cgroups to profile system activity of ansible and
          individual tasks, and display a recap at the end of the playbook execution
    notes:
        - Requires ansible to be run from within a cgroup, such as with
          C(cgexec -g cpuacct,memory,pids:ansible_profile ansible-playbook ...)
        - This cgroup should only be used by ansible to get accurate results
        - To create the cgroup, first use a command such as
          C(sudo cgcreate -a ec2-user:ec2-user -t ec2-user:ec2-user -g cpuacct,memory,pids:ansible_profile)
    options:
      control_group:
        required: True
        description: Name of cgroups control group
        env:
          - name: CGROUP_CONTROL_GROUP
        ini:
          - section: callback_cgroup_perf_recap
            key: control_group
      cpu_poll_interval:
        description: Interval between CPU polling for determining CPU usage. A lower value may produce inaccurate
                     results, a higher value may not be short enough to collect results for short tasks.
        default: 0.25
        type: float
        env:
          - name: CGROUP_CPU_POLL_INTERVAL
        ini:
          - section: callback_cgroup_perf_recap
            key: cpu_poll_interval
      memory_poll_interval:
        description: Interval between memory polling for determining memory usage. A lower value may produce inaccurate
                     results, a higher value may not be short enough to collect results for short tasks.
        default: 0.25
        type: float
        env:
          - name: CGROUP_MEMORY_POLL_INTERVAL
        ini:
          - section: callback_cgroup_perf_recap
            key: memory_poll_interval
      pid_poll_interval:
        description: Interval between PID polling for determining PID count. A lower value may produce inaccurate
                     results, a higher value may not be short enough to collect results for short tasks.
        default: 0.25
        type: float
        env:
          - name: CGROUP_PID_POLL_INTERVAL
        ini:
          - section: callback_cgroup_perf_recap
            key: pid_poll_interval
      display_recap:
        description: Controls whether the recap is printed at the end, useful if you will automatically
                     process the output files
        env:
          - name: CGROUP_DISPLAY_RECAP
        ini:
          - section: callback_cgroup_perf_recap
            key: display_recap
        type: bool
        default: true
      file_name_format:
        description: Format of filename. Accepts C(%(counter)s), C(%(task_uuid)s),
                     C(%(feature)s), C(%(ext)s). Defaults to C(%(feature)s.%(ext)s) when C(file_per_task) is C(False)
                     and C(%(counter)s-%(task_uuid)s-%(feature)s.%(ext)s) when C(True)
        env:
          - name: CGROUP_FILE_NAME_FORMAT
        ini:
          - section: callback_cgroup_perf_recap
            key: file_name_format
        type: str
        default: '%(feature)s.%(ext)s'
      output_dir:
        description: Output directory for files containing recorded performance readings. If the value contains a
                     single %s, the start time of the playbook run will be inserted in that space. Only the deepest
                     level directory will be created if it does not exist, parent directories will not be created.
        type: path
        default: /tmp/ansible-perf-%s
        env:
          - name: CGROUP_OUTPUT_DIR
        ini:
          - section: callback_cgroup_perf_recap
            key: output_dir
      output_format:
        description: Output format, either CSV or JSON-seq
        env:
          - name: CGROUP_OUTPUT_FORMAT
        ini:
          - section: callback_cgroup_perf_recap
            key: output_format
        type: str
        default: csv
        choices:
          - csv
          - json
      file_per_task:
        description: When set as C(True) along with C(write_files), this callback will write 1 file per task
                     instead of 1 file for the entire playbook run
        env:
          - name: CGROUP_FILE_PER_TASK
        ini:
          - section: callback_cgroup_perf_recap
            key: file_per_task
        type: bool
        default: False
      write_files:
        description: Dictates whether files will be written containing performance readings
        env:
          - name: CGROUP_WRITE_FILES
        ini:
          - section: callback_cgroup_perf_recap
            key: write_files
        type: bool
        default: false
'''

import csv
import datetime
import json
import os
import time
import threading

from abc import ABCMeta, abstractmethod

from functools import partial

from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.parsing.ajson import AnsibleJSONEncoder
from ansible.plugins.callback import CallbackBase


RS = '\x1e'  # RECORD SEPARATOR
LF = '\x0a'  # LINE FEED


def dict_fromkeys(keys, default=None):
    d = {}
    for key in keys:
        d[key] = default() if callable(default) else default
    return d


class BaseProf(threading.Thread, metaclass=ABCMeta):
    def __init__(self, path, obj=None, writer=None):
        threading.Thread.__init__(self)  # pylint: disable=non-parent-init-called
        self.obj = obj
        self.path = path
        self.max = 0
        self.running = True
        self.writer = writer

    def run(self):
        while self.running:
            self.poll()

    @abstractmethod
    def poll(self):
        pass


class MemoryProf(BaseProf):
    """Python thread for recording memory usage"""
    def __init__(self, path, poll_interval=0.25, obj=None, writer=None):
        super(MemoryProf, self).__init__(path, obj=obj, writer=writer)
        self._poll_interval = poll_interval

    def poll(self):
        with open(self.path) as f:
            val = int(f.read().strip()) / 1024**2
        if val > self.max:
            self.max = val
        if self.writer:
            try:
                self.writer(time.time(), self.obj.get_name(), self.obj._uuid, val)
            except ValueError:
                # We may be profiling after the playbook has ended
                self.running = False
        time.sleep(self._poll_interval)


class CpuProf(BaseProf):
    def __init__(self, path, poll_interval=0.25, obj=None, writer=None):
        super(CpuProf, self).__init__(path, obj=obj, writer=writer)
        self._poll_interval = poll_interval

    def poll(self):
        with open(self.path) as f:
            start_time = time.time() * 1000**2
            start_usage = int(f.read().strip()) / 1000
        time.sleep(self._poll_interval)
        with open(self.path) as f:
            end_time = time.time() * 1000**2
            end_usage = int(f.read().strip()) / 1000
        val = (end_usage - start_usage) / (end_time - start_time) * 100
        if val > self.max:
            self.max = val
        if self.writer:
            try:
                self.writer(time.time(), self.obj.get_name(), self.obj._uuid, val)
            except ValueError:
                # We may be profiling after the playbook has ended
                self.running = False


class PidsProf(BaseProf):
    def __init__(self, path, poll_interval=0.25, obj=None, writer=None):
        super(PidsProf, self).__init__(path, obj=obj, writer=writer)
        self._poll_interval = poll_interval

    def poll(self):
        with open(self.path) as f:
            val = int(f.read().strip())
        if val > self.max:
            self.max = val
        if self.writer:
            try:
                self.writer(time.time(), self.obj.get_name(), self.obj._uuid, val)
            except ValueError:
                # We may be profiling after the playbook has ended
                self.running = False
        time.sleep(self._poll_interval)


def csv_writer(writer, timestamp, task_name, task_uuid, value):
    writer.writerow([timestamp, task_name, task_uuid, value])


def json_writer(writer, timestamp, task_name, task_uuid, value):
    data = {
        'timestamp': timestamp,
        'task_name': task_name,
        'task_uuid': task_uuid,
        'value': value,
    }
    writer.write('%s%s%s' % (RS, json.dumps(data, cls=AnsibleJSONEncoder), LF))


class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'ansible.posix.cgroup_perf_recap'
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self, display=None):
        super(CallbackModule, self).__init__(display)

        self._features = ('memory', 'cpu', 'pids')

        self._units = {
            'memory': 'MB',
            'cpu': '%',
            'pids': '',
        }

        self.task_results = dict_fromkeys(self._features, default=list)
        self._profilers = dict.fromkeys(self._features)
        self._files = dict.fromkeys(self._features)
        self._writers = dict.fromkeys(self._features)

        self._file_per_task = False
        self._counter = 0
        self.write_files = False

    def _open_files(self, task_uuid=None):
        output_format = self._output_format
        output_dir = self._output_dir

        for feature in self._features:
            data = {
                b'counter': to_bytes(self._counter),
                b'task_uuid': to_bytes(task_uuid),
                b'feature': to_bytes(feature),
                b'ext': to_bytes(output_format)
            }

            if self._files.get(feature):
                try:
                    self._files[feature].close()
                except Exception:
                    pass

            if self.write_files:
                filename = self._file_name_format % data

                self._files[feature] = open(os.path.join(output_dir, filename), 'w+')
                if output_format == b'csv':
                    self._writers[feature] = partial(csv_writer, csv.writer(self._files[feature]))
                elif output_format == b'json':
                    self._writers[feature] = partial(json_writer, self._files[feature])

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super(CallbackModule, self).set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        cpu_poll_interval = self.get_option('cpu_poll_interval')
        memory_poll_interval = self.get_option('memory_poll_interval')
        pid_poll_interval = self.get_option('pid_poll_interval')
        self._display_recap = self.get_option('display_recap')

        control_group = to_bytes(self.get_option('control_group'), errors='surrogate_or_strict')
        self.mem_max_file = b'/sys/fs/cgroup/memory/%s/memory.max_usage_in_bytes' % control_group
        mem_current_file = b'/sys/fs/cgroup/memory/%s/memory.usage_in_bytes' % control_group
        cpu_usage_file = b'/sys/fs/cgroup/cpuacct/%s/cpuacct.usage' % control_group
        pid_current_file = b'/sys/fs/cgroup/pids/%s/pids.current' % control_group

        for path in (self.mem_max_file, mem_current_file, cpu_usage_file, pid_current_file):
            try:
                with open(path) as f:
                    pass
            except Exception as e:
                self._display.warning(
                    u'Cannot open %s for reading (%s). Disabling %s' % (to_text(path), to_text(e), self.CALLBACK_NAME)
                )
                self.disabled = True
                return

        try:
            with open(self.mem_max_file, 'w+') as f:
                f.write('0')
        except Exception as e:
            self._display.warning(
                u'Unable to reset max memory value in %s: %s' % (to_text(self.mem_max_file), to_text(e))
            )
            self.disabled = True
            return

        try:
            with open(cpu_usage_file, 'w+') as f:
                f.write('0')
        except Exception as e:
            self._display.warning(
                u'Unable to reset CPU usage value in %s: %s' % (to_text(cpu_usage_file), to_text(e))
            )
            self.disabled = True
            return

        self._profiler_map = {
            'memory': partial(MemoryProf, mem_current_file, poll_interval=memory_poll_interval),
            'cpu': partial(CpuProf, cpu_usage_file, poll_interval=cpu_poll_interval),
            'pids': partial(PidsProf, pid_current_file, poll_interval=pid_poll_interval),
        }

        self.write_files = self.get_option('write_files')
        file_per_task = self.get_option('file_per_task')
        self._output_format = to_bytes(self.get_option('output_format'))
        output_dir = to_bytes(self.get_option('output_dir'), errors='surrogate_or_strict')
        try:
            output_dir %= to_bytes(datetime.datetime.now().isoformat())
        except TypeError:
            pass

        self._output_dir = output_dir

        file_name_format = to_bytes(self.get_option('file_name_format'))

        if self.write_files:
            if file_per_task:
                self._file_per_task = True
                if file_name_format == b'%(feature)s.%(ext)s':
                    file_name_format = b'%(counter)s-%(task_uuid)s-%(feature)s.%(ext)s'
            else:
                file_name_format = to_bytes(self.get_option('file_name_format'))

            self._file_name_format = file_name_format

            if not os.path.exists(output_dir):
                try:
                    os.mkdir(output_dir)
                except Exception as e:
                    self._display.warning(
                        u'Could not create the output directory at %s: %s' % (to_text(output_dir), to_text(e))
                    )
                    self.disabled = True
                    return

            if not self._file_per_task:
                self._open_files()

    def _profile(self, obj=None):
        prev_task = None
        results = dict.fromkeys(self._features)
        if not obj or self._file_per_task:
            for dummy, f in self._files.items():
                if f is None:
                    continue
                try:
                    f.close()
                except Exception:
                    pass

        try:
            for name, prof in self._profilers.items():
                prof.running = False

            for name, prof in self._profilers.items():
                results[name] = prof.max
            prev_task = prof.obj
        except AttributeError:
            pass

        for name, result in results.items():
            if result is not None:
                try:
                    self.task_results[name].append((prev_task, result))
                except ValueError:
                    pass

        if obj is not None:
            if self._file_per_task or self._counter == 0:
                self._open_files(task_uuid=obj._uuid)

            for feature in self._features:
                self._profilers[feature] = self._profiler_map[feature](obj=obj, writer=self._writers[feature])
                self._profilers[feature].start()

            self._counter += 1

    def v2_playbook_on_task_start(self, task, is_conditional):
        self._profile(task)

    def v2_playbook_on_stats(self, stats):
        self._profile()

        if not self._display_recap:
            return

        with open(self.mem_max_file) as f:
            max_results = int(f.read().strip()) / 1024 / 1024

        self._display.banner('CGROUP PERF RECAP')
        self._display.display('Memory Execution Maximum: %0.2fMB\n' % max_results)
        for name, data in self.task_results.items():
            if name == 'memory':
                continue
            try:
                self._display.display(
                    '%s Execution Maximum: %0.2f%s\n' % (name, max((t[1] for t in data)), self._units[name])
                )
            except Exception as e:
                self._display.display('%s profiling error: no results collected: %s\n' % (name, e))

        self._display.display('\n')

        for name, data in self.task_results.items():
            if data:
                self._display.display('%s:\n' % name)
            for task, value in data:
                self._display.display('%s (%s): %0.2f%s' % (task.get_name(), task._uuid, value, self._units[name]))
            self._display.display('\n')



# Copyright (c) 2015, Filipe Niero Felisbino <filipenf@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
#
# contributed by Kelly Brazil <kellyjonbrazil@gmail.com>

from __future__ import annotations

DOCUMENTATION = r"""
name: jc
short_description: Convert output of many shell commands and file-types to JSON
version_added: 1.1.0
author: Kelly Brazil (@kellyjonbrazil)
description:
  - Convert output of many shell commands and file-types to JSON.
  - Uses the L(jc library,https://github.com/kellyjonbrazil/jc).
positional: parser
options:
  _input:
    description: The data to convert.
    type: string
    required: true
  parser:
    description:
      - The correct parser for the input data.
      - For example V(ifconfig).
      - 'Note: use underscores instead of dashes (if any) in the parser module name.'
      - See U(https://github.com/kellyjonbrazil/jc#parsers) for the latest list of parsers.
    type: string
    required: true
  quiet:
    description: Set to V(false) to not suppress warnings.
    type: boolean
    default: true
  raw:
    description: Set to V(true) to return pre-processed JSON.
    type: boolean
    default: false
requirements:
  - jc installed as a Python library (U(https://pypi.org/project/jc/))
"""

EXAMPLES = r"""
- name: Install the prereqs of the jc filter (jc Python package) on the Ansible controller
  delegate_to: localhost
  ansible.builtin.pip:
    name: jc
    state: present

- name: Run command
  ansible.builtin.command: uname -a
  register: result

- name: Convert command's result to JSON
  ansible.builtin.debug:
    msg: "{{ result.stdout | community.general.jc('uname') }}"
  # Possible output:
  #
  # "msg": {
  #   "hardware_platform": "x86_64",
  #   "kernel_name": "Linux",
  #   "kernel_release": "4.15.0-112-generic",
  #   "kernel_version": "#113-Ubuntu SMP Thu Jul 9 23:41:39 UTC 2020",
  #   "machine": "x86_64",
  #   "node_name": "kbrazil-ubuntu",
  #   "operating_system": "GNU/Linux",
  #   "processor": "x86_64"
  # }
"""

RETURN = r"""
_value:
  description: The processed output.
  type: any
"""

import importlib

from ansible.errors import AnsibleError, AnsibleFilterError

try:
    import jc

    HAS_LIB = True
except ImportError:
    HAS_LIB = False


def jc_filter(data, parser, quiet=True, raw=False):
    """Convert returned command output to JSON using the JC library

    Arguments:

        parser      required    (string) the correct parser for the input data (e.g. 'ifconfig')
                                see https://github.com/kellyjonbrazil/jc#parsers for latest list of parsers.
        quiet       optional    (bool) True to suppress warning messages (default is True)
        raw         optional    (bool) True to return pre-processed JSON (default is False)

    Returns:

        dictionary or list of dictionaries

    Example:
        - name: run date command
          hosts: ubuntu
          tasks:
          - name: install the prereqs of the jc filter (jc Python package) on the Ansible controller
            delegate_to: localhost
            ansible.builtin.pip:
              name: jc
              state: present
          - ansible.builtin.shell: date
            register: result
          - ansible.builtin.set_fact:
              myvar: "{{ result.stdout | community.general.jc('date') }}"
          - ansible.builtin.debug:
              msg: "{{ myvar }}"

        produces:

        ok: [192.168.1.239] => {
            "msg": {
                "day": 9,
                "hour": 22,
                "minute": 6,
                "month": "Aug",
                "month_num": 8,
                "second": 22,
                "timezone": "UTC",
                "weekday": "Sun",
                "weekday_num": 1,
                "year": 2020
            }
        }
    """

    if not HAS_LIB:
        raise AnsibleError(
            'You need to install "jc" as a Python library on the Ansible controller prior to running jc filter'
        )

    try:
        # new API (jc v1.18.0 and higher) allows use of plugin parsers
        if hasattr(jc, "parse"):
            return jc.parse(parser, data, quiet=quiet, raw=raw)

        # old API (jc v1.17.7 and lower)
        else:
            jc_parser = importlib.import_module(f"jc.parsers.{parser}")
            return jc_parser.parse(data, quiet=quiet, raw=raw)

    except Exception as e:
        raise AnsibleFilterError(f"Error in jc filter plugin: {e}") from e


class FilterModule:
    """Query filter"""

    def filters(self):
        return {
            "jc": jc_filter,
        }



#!/usr/bin/python
#
# Copyright (C) 2019 Huawei
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

###############################################################################
# Documentation
###############################################################################

DOCUMENTATION = r"""
module: hwc_evs_disk
description:
  - Block storage management.
short_description: Creates a resource of Evs/Disk in Huawei Cloud
version_added: '0.2.0'
author: Huawei Inc. (@huaweicloud)
requirements:
  - keystoneauth1 >= 3.6.0
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether the given object should exist in Huawei Cloud.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  timeouts:
    description:
      - The timeouts for each operations.
    type: dict
    default: {}
    suboptions:
      create:
        description:
          - The timeouts for create operation.
        type: str
        default: '30m'
      update:
        description:
          - The timeouts for update operation.
        type: str
        default: '30m'
      delete:
        description:
          - The timeouts for delete operation.
        type: str
        default: '30m'
  availability_zone:
    description:
      - Specifies the AZ where you want to create the disk.
    type: str
    required: true
  name:
    description:
      - Specifies the disk name. The value can contain a maximum of 255 bytes.
    type: str
    required: true
  volume_type:
    description:
      - Specifies the disk type. Currently, the value can be SSD, SAS, or SATA.
      - SSD specifies the ultra-high I/O disk type.
      - SAS specifies the high I/O disk type.
      - SATA specifies the common I/O disk type.
      - If the specified disk type is not available in the AZ, the disk creation fails. If the EVS disk is created from a
        snapshot, the volume_type field must be the same as that of the snapshot's source disk.
    type: str
    required: true
  backup_id:
    description:
      - Specifies the ID of the backup that can be used to create a disk. This parameter is mandatory when you use a backup
        to create the disk.
    type: str
  description:
    description:
      - Specifies the disk description. The value can contain a maximum of 255 bytes.
    type: str
  enable_full_clone:
    description:
      - If the disk is created from a snapshot and linked cloning needs to be used, set this parameter to True.
    type: bool
  enable_scsi:
    description:
      - If this parameter is set to V(true), the disk device type is SCSI, which allows ECS OSs to directly access underlying
        storage media. SCSI reservation command is supported. If this parameter is set to V(false), the disk device type is
        VBD, which supports only simple SCSI read/write commands.
      - If parameter enable_share is set to True and this parameter is not specified, shared SCSI disks are created. SCSI
        EVS disks cannot be created from backups, which means that this parameter cannot be True if backup_id has been specified.
    type: bool
  enable_share:
    description:
      - Specifies whether the disk is shareable. The default value is False.
    type: bool
  encryption_id:
    description:
      - Specifies the encryption ID. The length of it fixes at 36 bytes.
    type: str
  enterprise_project_id:
    description:
      - Specifies the enterprise project ID. This ID is associated with the disk during the disk creation. If it is not specified,
        the disk is bound to the default enterprise project.
    type: str
  image_id:
    description:
      - Specifies the image ID. If this parameter is specified, the disk is created from an image. BMS system disks cannot
        be created from BMS images.
    type: str
  size:
    description:
      - Specifies the disk size, in GB. Its values are as follows, System disk 1 GB to 1024 GB, Data disk 10 GB to 32768 GB.
        This parameter is mandatory when you create an empty disk or use an image or a snapshot to create a disk. If you use
        an image or a snapshot to create a disk, the disk size must be greater than or equal to the image or snapshot size.
        This parameter is optional when you use a backup to create a disk. If this parameter is not specified, the disk size
        is equal to the backup size.
    type: int
  snapshot_id:
    description:
      - Specifies the snapshot ID. If this parameter is specified, the disk is created from a snapshot.
    type: str
extends_documentation_fragment:
  - community.general._hwc
  - community.general._attributes
"""

EXAMPLES = r"""
# test create disk
- name: Create a disk
  community.general.hwc_evs_disk:
    availability_zone: "cn-north-1a"
    name: "ansible_evs_disk_test"
    volume_type: "SATA"
    size: 10
"""

RETURN = r"""
availability_zone:
  description:
    - Specifies the AZ where you want to create the disk.
  type: str
  returned: success
name:
  description:
    - Specifies the disk name. The value can contain a maximum of 255 bytes.
  type: str
  returned: success
volume_type:
  description:
    - Specifies the disk type. Currently, the value can be SSD, SAS, or SATA.
    - SSD specifies the ultra-high I/O disk type.
    - SAS specifies the high I/O disk type.
    - SATA specifies the common I/O disk type.
    - If the specified disk type is not available in the AZ, the disk creation fails. If the EVS disk is created from a snapshot,
      the volume_type field must be the same as that of the snapshot's source disk.
  type: str
  returned: success
backup_id:
  description:
    - Specifies the ID of the backup that can be used to create a disk. This parameter is mandatory when you use a backup
      to create the disk.
  type: str
  returned: success
description:
  description:
    - Specifies the disk description. The value can contain a maximum of 255 bytes.
  type: str
  returned: success
enable_full_clone:
  description:
    - If the disk is created from a snapshot and linked cloning needs to be used, set this parameter to True.
  type: bool
  returned: success
enable_scsi:
  description:
    - If this parameter is set to V(true), the disk device type is SCSI, which allows ECS OSs to directly access underlying
      storage media. SCSI reservation command is supported. If this parameter is set to V(false), the disk device type is
      VBD, which supports only simple SCSI read/write commands.
    - If parameter enable_share is set to True and this parameter is not specified, shared SCSI disks are created. SCSI EVS
      disks cannot be created from backups, which means that this parameter cannot be True if backup_id has been specified.
  type: bool
  returned: success
enable_share:
  description:
    - Specifies whether the disk is shareable. The default value is False.
  type: bool
  returned: success
encryption_id:
  description:
    - Specifies the encryption ID. The length of it fixes at 36 bytes.
  type: str
  returned: success
enterprise_project_id:
  description:
    - Specifies the enterprise project ID. This ID is associated with the disk during the disk creation. If it is not specified,
      the disk is bound to the default enterprise project.
  type: str
  returned: success
image_id:
  description:
    - Specifies the image ID. If this parameter is specified, the disk is created from an image. BMS system disks cannot be
      created from BMS images.
  type: str
  returned: success
size:
  description:
    - Specifies the disk size, in GB. Its values are as follows, System disk 1 GB to 1024 GB, Data disk 10 GB to 32768 GB.
      This parameter is mandatory when you create an empty disk or use an image or a snapshot to create a disk. If you use
      an image or a snapshot to create a disk, the disk size must be greater than or equal to the image or snapshot size.
      This parameter is optional when you use a backup to create a disk. If this parameter is not specified, the disk size
      is equal to the backup size.
  type: int
  returned: success
snapshot_id:
  description:
    - Specifies the snapshot ID. If this parameter is specified, the disk is created from a snapshot.
  type: str
  returned: success
attachments:
  description:
    - Specifies the disk attachment information.
  type: list
  elements: dict
  returned: success
  contains:
    attached_at:
      description:
        - Specifies the time when the disk was attached. Time format is 'UTC YYYY-MM-DDTHH:MM:SS'.
      type: str
      returned: success
    attachment_id:
      description:
        - Specifies the ID of the attachment information.
      type: str
      returned: success
    device:
      description:
        - Specifies the device name.
      type: str
      returned: success
    server_id:
      description:
        - Specifies the ID of the server to which the disk is attached.
      type: str
      returned: success
backup_policy_id:
  description:
    - Specifies the backup policy ID.
  type: str
  returned: success
created_at:
  description:
    - Specifies the time when the disk was created. Time format is 'UTC YYYY-MM-DDTHH:MM:SS'.
  type: str
  returned: success
is_bootable:
  description:
    - Specifies whether the disk is bootable.
  type: bool
  returned: success
is_readonly:
  description:
    - Specifies whether the disk is read-only or read/write. True indicates that the disk is read-only. False indicates that
      the disk is read/write.
  type: bool
  returned: success
source_volume_id:
  description:
    - Specifies the source disk ID. This parameter has a value if the disk is created from a source disk.
  type: str
  returned: success
status:
  description:
    - Specifies the disk status.
  type: str
  returned: success
tags:
  description:
    - Specifies the disk tags.
  type: dict
  returned: success
"""

from ansible_collections.community.general.plugins.module_utils._hwc_utils import (
    Config,
    HwcClientException,
    HwcModule,
    are_different_dicts,
    build_path,
    get_region,
    is_empty_value,
    navigate_value,
    wait_to_finish,
)


def build_module():
    return HwcModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"], type="str"),
            timeouts=dict(
                type="dict",
                options=dict(
                    create=dict(default="30m", type="str"),
                    update=dict(default="30m", type="str"),
                    delete=dict(default="30m", type="str"),
                ),
                default=dict(),
            ),
            availability_zone=dict(type="str", required=True),
            name=dict(type="str", required=True),
            volume_type=dict(type="str", required=True),
            backup_id=dict(type="str"),
            description=dict(type="str"),
            enable_full_clone=dict(type="bool"),
            enable_scsi=dict(type="bool"),
            enable_share=dict(type="bool"),
            encryption_id=dict(type="str"),
            enterprise_project_id=dict(type="str"),
            image_id=dict(type="str"),
            size=dict(type="int"),
            snapshot_id=dict(type="str"),
        ),
        supports_check_mode=True,
    )


def main():
    """Main function"""

    module = build_module()
    config = Config(module, "evs")

    try:
        _init(config)
        is_exist = module.params.get("id")

        result = None
        changed = False
        if module.params["state"] == "present":
            if not is_exist:
                if not module.check_mode:
                    create(config)
                changed = True

            inputv = user_input_parameters(module)
            resp, array_index = read_resource(config)
            result = build_state(inputv, resp, array_index)
            set_readonly_options(inputv, result)
            if are_different_dicts(inputv, result):
                if not module.check_mode:
                    update(config, inputv, result)

                    inputv = user_input_parameters(module)
                    resp, array_index = read_resource(config)
                    result = build_state(inputv, resp, array_index)
                    set_readonly_options(inputv, result)
                    if are_different_dicts(inputv, result):
                        raise Exception("Update resource failed, some attributes are not updated")

                changed = True

            result["id"] = module.params.get("id")
        else:
            result = dict()
            if is_exist:
                if not module.check_mode:
                    delete(config)
                changed = True

    except Exception as ex:
        module.fail_json(msg=str(ex))

    else:
        result["changed"] = changed
        module.exit_json(**result)


def _init(config):
    module = config.module
    if module.params.get("id"):
        return

    v = search_resource(config)
    n = len(v)
    if n > 1:
        raise Exception(f"find more than one resources({', '.join([navigate_value(i, ['id']) for i in v])})")

    if n == 1:
        module.params["id"] = navigate_value(v[0], ["id"])


def user_input_parameters(module):
    return {
        "availability_zone": module.params.get("availability_zone"),
        "backup_id": module.params.get("backup_id"),
        "description": module.params.get("description"),
        "enable_full_clone": module.params.get("enable_full_clone"),
        "enable_scsi": module.params.get("enable_scsi"),
        "enable_share": module.params.get("enable_share"),
        "encryption_id": module.params.get("encryption_id"),
        "enterprise_project_id": module.params.get("enterprise_project_id"),
        "image_id": module.params.get("image_id"),
        "name": module.params.get("name"),
        "size": module.params.get("size"),
        "snapshot_id": module.params.get("snapshot_id"),
        "volume_type": module.params.get("volume_type"),
    }


def create(config):
    module = config.module
    client = config.client(get_region(module), "volumev3", "project")
    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    opts = user_input_parameters(module)
    opts["ansible_module"] = module

    params = build_create_parameters(opts)
    r = send_create_request(module, params, client)

    client1 = config.client(get_region(module), "volume", "project")
    client1.endpoint = client1.endpoint.replace("/v2/", "/v1/")
    obj = async_wait(config, r, client1, timeout)
    module.params["id"] = navigate_value(obj, ["entities", "volume_id"])


def update(config, expect_state, current_state):
    module = config.module
    expect_state["current_state"] = current_state
    current_state["current_state"] = current_state
    client = config.client(get_region(module), "evs", "project")
    timeout = 60 * int(module.params["timeouts"]["update"].rstrip("m"))

    params = build_update_parameters(expect_state)
    params1 = build_update_parameters(current_state)
    if params and are_different_dicts(params, params1):
        send_update_request(module, params, client)

    params = build_extend_disk_parameters(expect_state)
    params1 = build_extend_disk_parameters(current_state)
    if params and are_different_dicts(params, params1):
        client1 = config.client(get_region(module), "evsv2.1", "project")
        r = send_extend_disk_request(module, params, client1)

        client1 = config.client(get_region(module), "volume", "project")
        client1.endpoint = client1.endpoint.replace("/v2/", "/v1/")
        async_wait(config, r, client1, timeout)


def delete(config):
    module = config.module
    client = config.client(get_region(module), "evs", "project")
    timeout = 60 * int(module.params["timeouts"]["delete"].rstrip("m"))

    r = send_delete_request(module, None, client)

    client = config.client(get_region(module), "volume", "project")
    client.endpoint = client.endpoint.replace("/v2/", "/v1/")
    async_wait(config, r, client, timeout)


def read_resource(config):
    module = config.module
    client = config.client(get_region(module), "volumev3", "project")

    res = {}

    r = send_read_request(module, client)
    res["read"] = fill_read_resp_body(r)

    return res, None


def build_state(opts, response, array_index):
    states = flatten_options(response, array_index)
    set_unreadable_options(opts, states)
    return states


def _build_query_link(opts):
    query_params = []

    v = navigate_value(opts, ["enable_share"])
    if v or v in [False, 0]:
        query_params.append(f"multiattach={str(v) if v else str(v).lower()}")

    v = navigate_value(opts, ["name"])
    if v or v in [False, 0]:
        query_params.append(f"name={str(v) if v else str(v).lower()}")

    v = navigate_value(opts, ["availability_zone"])
    if v or v in [False, 0]:
        query_params.append(f"availability_zone={str(v) if v else str(v).lower()}")

    query_link = "?limit=10&offset={start}"
    if query_params:
        query_link += f"&{'&'.join(query_params)}"

    return query_link


def search_resource(config):
    module = config.module
    client = config.client(get_region(module), "volumev3", "project")
    opts = user_input_parameters(module)
    name = module.params.get("name")
    query_link = _build_query_link(opts)
    link = f"os-vendor-volumes/detail{query_link}"

    result = []
    p = {"start": 0}
    while True:
        url = link.format(**p)
        r = send_list_request(module, client, url)
        if not r:
            break

        for item in r:
            if name == item.get("name"):
                result.append(item)

        if len(result) > 1:
            break

        p["start"] += len(r)

    return result


def build_create_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["availability_zone"], None)
    if not is_empty_value(v):
        params["availability_zone"] = v

    v = navigate_value(opts, ["backup_id"], None)
    if not is_empty_value(v):
        params["backup_id"] = v

    v = navigate_value(opts, ["description"], None)
    if not is_empty_value(v):
        params["description"] = v

    v = navigate_value(opts, ["enterprise_project_id"], None)
    if not is_empty_value(v):
        params["enterprise_project_id"] = v

    v = navigate_value(opts, ["image_id"], None)
    if not is_empty_value(v):
        params["imageRef"] = v

    v = expand_create_metadata(opts, None)
    if not is_empty_value(v):
        params["metadata"] = v

    v = navigate_value(opts, ["enable_share"], None)
    if not is_empty_value(v):
        params["multiattach"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = navigate_value(opts, ["size"], None)
    if not is_empty_value(v):
        params["size"] = v

    v = navigate_value(opts, ["snapshot_id"], None)
    if not is_empty_value(v):
        params["snapshot_id"] = v

    v = navigate_value(opts, ["volume_type"], None)
    if not is_empty_value(v):
        params["volume_type"] = v

    if not params:
        return params

    params = {"volume": params}

    return params


def expand_create_metadata(d, array_index):
    r = dict()

    v = navigate_value(d, ["encryption_id"], array_index)
    if not is_empty_value(v):
        r["__system__cmkid"] = v

    v = expand_create_metadata_system_encrypted(d, array_index)
    if not is_empty_value(v):
        r["__system__encrypted"] = v

    v = expand_create_metadata_full_clone(d, array_index)
    if not is_empty_value(v):
        r["full_clone"] = v

    v = expand_create_metadata_hw_passthrough(d, array_index)
    if not is_empty_value(v):
        r["hw:passthrough"] = v

    return r


def expand_create_metadata_system_encrypted(d, array_index):
    v = navigate_value(d, ["encryption_id"], array_index)
    return "1" if v else ""


def expand_create_metadata_full_clone(d, array_index):
    v = navigate_value(d, ["enable_full_clone"], array_index)
    return "0" if v else ""


def expand_create_metadata_hw_passthrough(d, array_index):
    v = navigate_value(d, ["enable_scsi"], array_index)
    if v is None:
        return v
    return "true" if v else "false"


def send_create_request(module, params, client):
    url = "cloudvolumes"
    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(create), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_update_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["description"], None)
    if v is not None:
        params["description"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    if not params:
        return params

    params = {"volume": params}

    return params


def send_update_request(module, params, client):
    url = build_path(module, "cloudvolumes/{id}")

    try:
        r = client.put(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(update), error: {ex}"
        module.fail_json(msg=msg)

    return r


def send_delete_request(module, params, client):
    url = build_path(module, "cloudvolumes/{id}")

    try:
        r = client.delete(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(delete), error: {ex}"
        module.fail_json(msg=msg)

    return r


def build_extend_disk_parameters(opts):
    params = dict()

    v = expand_extend_disk_os_extend(opts, None)
    if not is_empty_value(v):
        params["os-extend"] = v

    return params


def expand_extend_disk_os_extend(d, array_index):
    r = dict()

    v = navigate_value(d, ["size"], array_index)
    if not is_empty_value(v):
        r["new_size"] = v

    return r


def send_extend_disk_request(module, params, client):
    url = build_path(module, "cloudvolumes/{id}/action")

    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(extend_disk), error: {ex}"
        module.fail_json(msg=msg)

    return r


def async_wait(config, result, client, timeout):
    module = config.module

    path_parameters = {
        "job_id": ["job_id"],
    }
    data = {key: navigate_value(result, path) for key, path in path_parameters.items()}

    url = build_path(module, "jobs/{job_id}", data)

    def _query_status():
        r = None
        try:
            r = client.get(url, timeout=timeout)
        except HwcClientException:
            return None, ""

        try:
            s = navigate_value(r, ["status"])
            return r, s
        except Exception:
            return None, ""

    try:
        return wait_to_finish(["SUCCESS"], ["RUNNING", "INIT"], _query_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_evs_disk): error waiting to be done, error= {ex}")


def send_read_request(module, client):
    url = build_path(module, "os-vendor-volumes/{id}")

    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(read), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["volume"], None)


def fill_read_resp_body(body):
    result = dict()

    v = fill_read_resp_attachments(body.get("attachments"))
    result["attachments"] = v

    result["availability_zone"] = body.get("availability_zone")

    result["bootable"] = body.get("bootable")

    result["created_at"] = body.get("created_at")

    result["description"] = body.get("description")

    result["enterprise_project_id"] = body.get("enterprise_project_id")

    result["id"] = body.get("id")

    v = fill_read_resp_metadata(body.get("metadata"))
    result["metadata"] = v

    result["multiattach"] = body.get("multiattach")

    result["name"] = body.get("name")

    result["size"] = body.get("size")

    result["snapshot_id"] = body.get("snapshot_id")

    result["source_volid"] = body.get("source_volid")

    result["status"] = body.get("status")

    result["tags"] = body.get("tags")

    v = fill_read_resp_volume_image_metadata(body.get("volume_image_metadata"))
    result["volume_image_metadata"] = v

    result["volume_type"] = body.get("volume_type")

    return result


def fill_read_resp_attachments(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["attached_at"] = item.get("attached_at")

        val["attachment_id"] = item.get("attachment_id")

        val["device"] = item.get("device")

        val["server_id"] = item.get("server_id")

        result.append(val)

    return result


def fill_read_resp_metadata(value):
    if not value:
        return None

    result = dict()

    result["__system__cmkid"] = value.get("__system__cmkid")

    result["attached_mode"] = value.get("attached_mode")

    result["full_clone"] = value.get("full_clone")

    result["hw:passthrough"] = value.get("hw:passthrough")

    result["policy"] = value.get("policy")

    result["readonly"] = value.get("readonly")

    return result


def fill_read_resp_volume_image_metadata(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


def flatten_options(response, array_index):
    r = dict()

    v = flatten_attachments(response, array_index)
    r["attachments"] = v

    v = navigate_value(response, ["read", "availability_zone"], array_index)
    r["availability_zone"] = v

    v = navigate_value(response, ["read", "metadata", "policy"], array_index)
    r["backup_policy_id"] = v

    v = navigate_value(response, ["read", "created_at"], array_index)
    r["created_at"] = v

    v = navigate_value(response, ["read", "description"], array_index)
    r["description"] = v

    v = flatten_enable_full_clone(response, array_index)
    r["enable_full_clone"] = v

    v = flatten_enable_scsi(response, array_index)
    r["enable_scsi"] = v

    v = navigate_value(response, ["read", "multiattach"], array_index)
    r["enable_share"] = v

    v = navigate_value(response, ["read", "metadata", "__system__cmkid"], array_index)
    r["encryption_id"] = v

    v = navigate_value(response, ["read", "enterprise_project_id"], array_index)
    r["enterprise_project_id"] = v

    v = navigate_value(response, ["read", "volume_image_metadata", "id"], array_index)
    r["image_id"] = v

    v = flatten_is_bootable(response, array_index)
    r["is_bootable"] = v

    v = flatten_is_readonly(response, array_index)
    r["is_readonly"] = v

    v = navigate_value(response, ["read", "name"], array_index)
    r["name"] = v

    v = navigate_value(response, ["read", "size"], array_index)
    r["size"] = v

    v = navigate_value(response, ["read", "snapshot_id"], array_index)
    r["snapshot_id"] = v

    v = navigate_value(response, ["read", "source_volid"], array_index)
    r["source_volume_id"] = v

    v = navigate_value(response, ["read", "status"], array_index)
    r["status"] = v

    v = navigate_value(response, ["read", "tags"], array_index)
    r["tags"] = v

    v = navigate_value(response, ["read", "volume_type"], array_index)
    r["volume_type"] = v

    return r


def flatten_attachments(d, array_index):
    v = navigate_value(d, ["read", "attachments"], array_index)
    if not v:
        return None
    n = len(v)
    result = []

    new_ai = dict()
    if array_index:
        new_ai.update(array_index)

    for i in range(n):
        new_ai["read.attachments"] = i

        val = dict()

        v = navigate_value(d, ["read", "attachments", "attached_at"], new_ai)
        val["attached_at"] = v

        v = navigate_value(d, ["read", "attachments", "attachment_id"], new_ai)
        val["attachment_id"] = v

        v = navigate_value(d, ["read", "attachments", "device"], new_ai)
        val["device"] = v

        v = navigate_value(d, ["read", "attachments", "server_id"], new_ai)
        val["server_id"] = v

        for v in val.values():
            if v is not None:
                result.append(val)
                break

    return result if result else None


def flatten_enable_full_clone(d, array_index):
    v = navigate_value(d, ["read", "metadata", "full_clone"], array_index)
    if v is None:
        return v
    return True if v == "0" else False


def flatten_enable_scsi(d, array_index):
    v = navigate_value(d, ["read", "metadata", "hw:passthrough"], array_index)
    if v is None:
        return v
    return True if v in ["true", "True"] else False


def flatten_is_bootable(d, array_index):
    v = navigate_value(d, ["read", "bootable"], array_index)
    if v is None:
        return v
    return True if v in ["true", "True"] else False


def flatten_is_readonly(d, array_index):
    v = navigate_value(d, ["read", "metadata", "readonly"], array_index)
    if v is None:
        return v
    return True if v in ["true", "True"] else False


def set_unreadable_options(opts, states):
    states["backup_id"] = opts.get("backup_id")


def set_readonly_options(opts, states):
    opts["attachments"] = states.get("attachments")

    opts["backup_policy_id"] = states.get("backup_policy_id")

    opts["created_at"] = states.get("created_at")

    opts["is_bootable"] = states.get("is_bootable")

    opts["is_readonly"] = states.get("is_readonly")

    opts["source_volume_id"] = states.get("source_volume_id")

    opts["status"] = states.get("status")

    opts["tags"] = states.get("tags")


def send_list_request(module, client, url):
    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_evs_disk): error running api(list), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["volumes"], None)


def expand_list_metadata(d, array_index):
    r = dict()

    v = navigate_value(d, ["encryption_id"], array_index)
    r["__system__cmkid"] = v

    r["attached_mode"] = None

    v = navigate_value(d, ["enable_full_clone"], array_index)
    r["full_clone"] = v

    v = navigate_value(d, ["enable_scsi"], array_index)
    r["hw:passthrough"] = v

    r["policy"] = None

    r["readonly"] = None

    for v in r.values():
        if v is not None:
            return r
    return None


def expand_list_volume_image_metadata(d, array_index):
    r = dict()

    v = navigate_value(d, ["image_id"], array_index)
    r["id"] = v

    for v in r.values():
        if v is not None:
            return r
    return None


def fill_list_resp_body(body):
    result = dict()

    v = fill_list_resp_attachments(body.get("attachments"))
    result["attachments"] = v

    result["availability_zone"] = body.get("availability_zone")

    result["bootable"] = body.get("bootable")

    result["created_at"] = body.get("created_at")

    result["description"] = body.get("description")

    result["enterprise_project_id"] = body.get("enterprise_project_id")

    result["id"] = body.get("id")

    v = fill_list_resp_metadata(body.get("metadata"))
    result["metadata"] = v

    result["multiattach"] = body.get("multiattach")

    result["name"] = body.get("name")

    result["size"] = body.get("size")

    result["snapshot_id"] = body.get("snapshot_id")

    result["source_volid"] = body.get("source_volid")

    result["status"] = body.get("status")

    result["tags"] = body.get("tags")

    v = fill_list_resp_volume_image_metadata(body.get("volume_image_metadata"))
    result["volume_image_metadata"] = v

    result["volume_type"] = body.get("volume_type")

    return result


def fill_list_resp_attachments(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["attached_at"] = item.get("attached_at")

        val["attachment_id"] = item.get("attachment_id")

        val["device"] = item.get("device")

        val["server_id"] = item.get("server_id")

        result.append(val)

    return result


def fill_list_resp_metadata(value):
    if not value:
        return None

    result = dict()

    result["__system__cmkid"] = value.get("__system__cmkid")

    result["attached_mode"] = value.get("attached_mode")

    result["full_clone"] = value.get("full_clone")

    result["hw:passthrough"] = value.get("hw:passthrough")

    result["policy"] = value.get("policy")

    result["readonly"] = value.get("readonly")

    return result


def fill_list_resp_volume_image_metadata(value):
    if not value:
        return None

    result = dict()

    result["id"] = value.get("id")

    return result


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr bgp_templates fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.argspec.bgp_templates.bgp_templates import (
    Bgp_templatesArgs,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.bgp_templates import (
    Bgp_templatesTemplate,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    flatten_config,
)


class Bgp_templatesFacts(object):
    """The iosxr bgp_templates facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Bgp_templatesArgs.argument_spec

    def get_config(self, connection):
        return connection.get("show running-config router bgp")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for Bgp_templates network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        objs = []

        if not data:
            data = self.get_config(connection)
        data = flatten_config(data, "neighbor-group")
        # parse native config using the Bgp_templates template
        bgp_templates_parser = Bgp_templatesTemplate(lines=data.splitlines(), module=self._module)
        objs = bgp_templates_parser.parse()
        if objs:
            objs["neighbor"] = self._post_parse(objs).get(
                "neighbor",
                [],
            )
        ansible_facts["ansible_network_resources"].pop("bgp_templates", None)

        params = utils.remove_empties(
            bgp_templates_parser.validate_config(self.argument_spec, {"config": objs}, redact=True),
        )

        facts["bgp_templates"] = params.get("config", {})
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts

    def _post_parse(self, data):
        """Converts the intermediate data structure
            to valid format as per argspec.
        :param obj: dict
        """
        if "neighbor" in data:
            data["neighbor"] = sorted(
                list(data["neighbor"].values()),
                key=lambda k, s="name": k[s],
            )
            for nbrg in data["neighbor"]:
                if nbrg.get("address_family"):
                    nbrg["address_family"] = list(nbrg["address_family"].values())
        return data



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Interfaces parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class InterfacesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(InterfacesTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            'name': 'interface',
            'getval': re.compile(
                r'''
              ^interface\s
              (?P<name>\S+)$''', re.VERBOSE,
            ),
            'setval': 'interface {{ name }}',
            'result': {
                '{{ name }}': {
                    'name': '{{ name }}',
                    'enabled': True,
                },
            },
            'shared': True,
        },
        {
            "name": "description",
            "getval": re.compile(
                r"""
                \s+description\s(?P<description>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "description {{ description }}",
            "result": {
                '{{ name }}': {
                    'description': "'{{ description }}'",
                },
            },
        },
        {
            "name": "enabled",
            "getval": re.compile(
                r"""
                (?P<negate>\sno)?
                (?P<shutdown>\sshutdown)
                $""", re.VERBOSE,
            ),
            "setval": "shutdown",
            "result": {
                '{{ name }}': {
                    'enabled': "{{ False if shutdown is defined and negate is not defined else True }}",
                },
            },
        },
        {
            "name": "mac_address",
            "getval": re.compile(
                r"""
                \s+mac-address
                (\s(?P<mac_address>.+))
                $""", re.VERBOSE,
            ),
            "setval": "mac-address {{ mac_address }}",
            "result": {
                '{{ name }}': {
                    'mac_address': "{{ mac_address }}",
                },
            },
        },
        {
            "name": "service_policy.input",
            "getval": re.compile(
                r"""
                \s+service-policy\sinput
                (\s(?P<enabled>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy input {{ service_policy.input }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "input": "{{ enabled }}",
                    },
                },
            },
        },
        {
            "name": "service_policy.output",
            "getval": re.compile(
                r"""
                \s+service-policy\soutput
                (\s(?P<enabled>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy output {{ service_policy.output }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "output": "{{ enabled }}",
                    },
                },
            },
        },
        {
            "name": "service_policy.type_options.access_control",
            "getval": re.compile(
                r"""
                \s+service-policy\stype\saccess-control
                (\sinput\s(?P<input>\S+))?
                (\soutput\s(?P<output>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy type access-control"
                      "{{ service_policy.type_options.access_control.input if service_policy.type_options.access_control.input is defined else '' }}"
                      "{{ service_policy.type_options.access_control.output if service_policy.type_options.access_control.output is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "type_options": {
                            "access_control": {
                                "input": "{{ input }}",
                                "output": "{{ output }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "service_policy.type_options.epbr",
            "getval": re.compile(
                r"""
                \s+service-policy\stype\sepbr
                (\sinput\s(?P<input>\S+))?
                (\soutput\s(?P<output>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy type epbr"
                      "{{ service_policy.type_options.epbr.input if service_policy.type_options.epbr.input is defined else '' }}"
                      "{{ service_policy.type_options.epbr.output if service_policy.type_options.epbr.output is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "type_options": {
                            "epbr": {
                                "input": "{{ input }}",
                                "output": "{{ output }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "service_policy.type_options.nwpi",
            "getval": re.compile(
                r"""
                \s+service-policy\stype\snwpi
                (\sinput\s(?P<input>\S+))?
                (\soutput\s(?P<output>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy type nwpi"
                      "{{ service_policy.type_options.nwpi.input if service_policy.type_options.nwpi.input is defined else '' }}"
                      "{{ service_policy.type_options.nwpi.output if service_policy.type_options.nwpi.output is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "type_options": {
                            "nwpi": {
                                "input": "{{ input }}",
                                "output": "{{ output }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "service_policy.type_options.packet_service",
            "getval": re.compile(
                r"""
                \s+service-policy\stype\spacket-service
                (\sinput\s(?P<input>\S+))?
                (\soutput\s(?P<output>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy type packet-service"
                      "{{ service_policy.type_options.packet_service.input if service_policy.type_options.packet_service.input is defined else '' }}"
                      "{{ service_policy.type_options.packet_service.output if service_policy.type_options.packet_service.output is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "type_options": {
                            "packet_service": {
                                "input": "{{ input }}",
                                "output": "{{ output }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "service_policy.type_options.service_chain",
            "getval": re.compile(
                r"""
                \s+service-policy\stype\sservice-chain
                (\sinput\s(?P<input>\S+))?
                (\soutput\s(?P<output>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "service-policy type service-chain"
                      "{{ service_policy.type_options.service_chain.input if service_policy.type_options.service_chain.input is defined else '' }}"
                      "{{ service_policy.type_options.service_chain.output if service_policy.type_options.service_chain.output is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "service_policy": {
                        "type_options": {
                            "service_chain": {
                                "input": "{{ input }}",
                                "output": "{{ output }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "logging.trunk_status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\strunk-status
                $""", re.VERBOSE,
            ),
            "setval": "logging event trunk-status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "trunk_status": True,
                    },
                },
            },
        },
        {
            "name": "logging.subif_link_status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\ssubif-link-status
                $""", re.VERBOSE,
            ),
            "setval": "logging event subif-link-status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "subif_link_status": True,
                    },
                },
            },
        },
        {
            "name": "logging.status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\sstatus
                $""", re.VERBOSE,
            ),
            "setval": "logging event status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "status": True,
                    },
                },
            },
        },
        {
            "name": "logging.spanning_tree",
            "getval": re.compile(
                r"""
                \s+logging\sevent\sspanning-tree
                $""", re.VERBOSE,
            ),
            "setval": "logging event spanning-tree",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "spanning_tree": True,
                    },
                },
            },
        },
        {
            "name": "logging.nfas_status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\snfas-status
                $""", re.VERBOSE,
            ),
            "setval": "logging event nfas-status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "nfas_status": True,
                    },
                },
            },
        },
        {
            "name": "logging.bundle_status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\sbundle-status
                $""", re.VERBOSE,
            ),
            "setval": "logging event bundle-status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "bundle_status": True,
                    },
                },
            },
        },
        {
            "name": "logging.link_status",
            "getval": re.compile(
                r"""
                \s+logging\sevent\slink-status
                $""", re.VERBOSE,
            ),
            "setval": "logging event link-status",
            "result": {
                "{{ name }}": {
                    "logging": {
                        "link_status": True,
                    },
                },
            },
        },
        {
            "name": "snmp.trap",
            "getval": re.compile(
                r"""
                \s+snmp\strap
                (\s(?P<ip>ip\sverify\sdrop-rate))?
                (\s(?P<link_status>link-status\spermit\sduplicates))?
                (\s(?P<mac_notification_added>mac-notification-added))?
                (\s(?P<mac_notification_removed>mac-notification-removed))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp trap"
                      "{{ ' ip verify drop-rate' if snmp.trap.ip is defined else '' }}"
                      "{{ ' link-status permit duplicates' if snmp.trap.link_status is defined else '' }}"
                      "{{ ' mac-notification-added' if snmp.trap.mac_notification_added is defined else '' }}"
                      "{{ ' mac-notification-removed' if snmp.trap.mac_notification_removed is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "snmp": {
                        "trap": {
                            "ip": "{{ not not ip }}",
                            "link_status": "{{ not not link_status }}",
                            "mac_notification_added": "{{ not not mac_notification_added }}",
                            "mac_notification_removed": "{{ not not mac_notification_removed }}",
                        },
                    },
                },
            },
        },
        {
            "name": "snmp.ifindex",
            "getval": re.compile(
                r"""
                \s+snmp\sifindex
                (\s(?P<clear>clear))?
                (\s(?P<persist>persist))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp ifindex"
                      "{{ ' clear' if snmp.ifindex.clear is defined else '' }}"
                      "{{ ' persist' if snmp.ifindex.persist is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "snmp": {
                        "ifindex": {
                            "clear": "{{ not not clear }}",
                            "persist": "{{ not not persist }}",
                        },
                    },
                },
            },
        },
        {  # only applicable for switches
            "name": "mode",
            "getval": re.compile(
                r"""
                (?P<negate>\sno)?
                (?P<switchport>\sswitchport)
                $""", re.VERBOSE,
            ),
            "setval": "switchport",
            "result": {
                '{{ name }}': {
                    'mode': "{{ 'layer2' if switchport is defined and negate is not defined else 'layer3' }}",
                },
            },
        },
        {
            "name": "speed",
            "getval": re.compile(
                r"""
                \s+speed\s(?P<speed>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "speed {{ speed|string }}",
            "result": {
                '{{ name }}': {
                    'speed': '{{ speed }}',
                },
            },
        },
        {
            "name": "mtu",
            "getval": re.compile(
                r"""
                \s+mtu\s(?P<mtu>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "mtu {{ mtu|string }}",
            "result": {
                '{{ name }}': {
                    'mtu': '{{ mtu }}',
                },
            },
        },
        {
            "name": "duplex",
            "getval": re.compile(
                r"""
                \s+duplex\s(?P<duplex>full|half|auto)
                $""", re.VERBOSE,
            ),
            "setval": "duplex {{ duplex }}",
            "result": {
                '{{ name }}': {
                    'duplex': '{{ duplex }}',
                },
            },
        },
        {
            "name": "template",
            "getval": re.compile(
                r"""
                \s+source\stemplate\s(?P<template>.+$)
                """, re.VERBOSE,
            ),
            "setval": "source template {{ template }}",
            "result": {
                '{{ name }}': {
                    'template': '{{ template }}',
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
# Copyright (c) 2024 Alexander Bakanovskii <skottttt228@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: krb_ticket
short_description: Kerberos utils for managing tickets
version_added: 10.0.0
description:
  - Manage Kerberos tickets with C(kinit), C(klist) and C(kdestroy) base utilities.
  - See U(https://web.mit.edu/kerberos/krb5-1.12/doc/user/user_commands/index.html) for reference.
author: "Alexander Bakanovskii (@abakanovskii)"
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  password:
    description:
      - Principal password.
      - It is required to specify O(password) or O(keytab_path).
    type: str
  principal:
    description:
      - The principal name.
      - If not set, the user running this module is used.
    type: str
  state:
    description:
      - The state of the Kerberos ticket.
      - V(present) is equivalent of C(kinit) command.
      - V(absent) is equivalent of C(kdestroy) command.
    type: str
    default: present
    choices: ["present", "absent"]
  kdestroy_all:
    description:
      - When O(state=absent) destroys all credential caches in collection.
      - Equivalent of running C(kdestroy -A).
    type: bool
  cache_name:
    description:
      - Use O(cache_name) as the ticket cache name and location.
      - If this option is not used, the default cache name and location are used.
      - The default credentials cache may vary between systems.
      - If not set the value of E(KRB5CCNAME) environment variable is used instead, its value is used to name the default
        ticket cache.
    type: str
  lifetime:
    description:
      - Requests a ticket with the lifetime, if the O(lifetime) is not specified, the default ticket lifetime is used.
      - Specifying a ticket lifetime longer than the maximum ticket lifetime (configured by each site) does not override the
        configured maximum ticket lifetime.
      - 'The value for O(lifetime) must be followed by one of the following suffixes: V(s) - seconds, V(m) - minutes, V(h)
        - hours, V(d) - days.'
      - You cannot mix units; a value of V(3h30m) results in an error.
      - See U(https://web.mit.edu/kerberos/krb5-1.12/doc/basic/date_format.html) for reference.
    type: str
  start_time:
    description:
      - Requests a postdated ticket.
      - Postdated tickets are issued with the invalid flag set, and need to be resubmitted to the KDC for validation before
        use.
      - O(start_time) specifies the duration of the delay before the ticket can become valid.
      - You can use absolute time formats, for example V(July 27, 2012 at 20:30) you would neet to set O(start_time=20120727203000).
      - You can also use time duration format similar to O(lifetime) or O(renewable).
      - See U(https://web.mit.edu/kerberos/krb5-1.12/doc/basic/date_format.html) for reference.
    type: str
  renewable:
    description:
      - Requests renewable tickets, with a total lifetime equal to O(renewable).
      - 'The value for O(renewable) must be followed by one of the following delimiters: V(s) - seconds, V(m) - minutes, V(h)
        - hours, V(d) - days.'
      - You cannot mix units; a value of V(3h30m) results in an error.
      - See U(https://web.mit.edu/kerberos/krb5-1.12/doc/basic/date_format.html) for reference.
    type: str
  forwardable:
    description:
      - Request forwardable or non-forwardable tickets.
    type: bool
  proxiable:
    description:
      - Request proxiable or non-proxiable tickets.
    type: bool
  address_restricted:
    description:
      - Request tickets restricted to the host's local address or non-restricted.
    type: bool
  anonymous:
    description:
      - Requests anonymous processing.
    type: bool
  canonicalization:
    description:
      - Requests canonicalization of the principal name, and allows the KDC to reply with a different client principal from
        the one requested.
    type: bool
  enterprise:
    description:
      - Treats the principal name as an enterprise name (implies the O(canonicalization) option).
    type: bool
  renewal:
    description:
      - Requests renewal of the ticket-granting ticket.
      - Note that an expired ticket cannot be renewed, even if the ticket is still within its renewable life.
    type: bool
  validate:
    description:
      - Requests that the ticket-granting ticket in the cache (with the invalid flag set) be passed to the KDC for validation.
      - If the ticket is within its requested time range, the cache is replaced with the validated ticket.
    type: bool
  keytab:
    description:
      - Requests a ticket, obtained from a key in the local host's keytab.
      - If O(keytab_path) is not specified it tries to use default client keytab path (C(-i) option).
    type: bool
  keytab_path:
    description:
      - Use when O(keytab=true) to specify path to a keytab file.
      - It is required to specify O(password) or O(keytab_path).
    type: path
requirements:
  - krb5-user and krb5-config packages
extends_documentation_fragment:
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Get Kerberos ticket using default principal
  community.general.krb_ticket:
    password: some_password

- name: Get Kerberos ticket using keytab
  community.general.krb_ticket:
    keytab: true
    keytab_path: /etc/ipa/file.keytab

- name: Get Kerberos ticket with a lifetime of 7 days
  community.general.krb_ticket:
    password: some_password
    lifetime: 7d

- name: Get Kerberos ticket with a starting time of July 2, 2024, 1:35:30 p.m.
  community.general.krb_ticket:
    password: some_password
    start_time: "240702133530"

- name: Get Kerberos ticket using principal name
  community.general.krb_ticket:
    password: some_password
    principal: admin

- name: Get Kerberos ticket using principal with realm
  community.general.krb_ticket:
    password: some_password
    principal: admin@IPA.TEST

- name: Check for existence by ticket cache
  community.general.krb_ticket:
    cache_name: KEYRING:persistent:0:0

- name: Make sure default ticket is destroyed
  community.general.krb_ticket:
    state: absent

- name: Make sure specific ticket destroyed by principal
  community.general.krb_ticket:
    state: absent
    principal: admin@IPA.TEST

- name: Make sure specific ticket destroyed by cache_name
  community.general.krb_ticket:
    state: absent
    cache_name: KEYRING:persistent:0:0

- name: Make sure all tickets are destroyed
  community.general.krb_ticket:
    state: absent
    kdestroy_all: true
"""

from ansible.module_utils.basic import AnsibleModule, env_fallback

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt


class IPAKeytab:
    def __init__(self, module, **kwargs):
        self.module = module
        self.password = kwargs["password"]
        self.principal = kwargs["principal"]
        self.state = kwargs["state"]
        self.kdestroy_all = kwargs["kdestroy_all"]
        self.cache_name = kwargs["cache_name"]
        self.start_time = kwargs["start_time"]
        self.renewable = kwargs["renewable"]
        self.forwardable = kwargs["forwardable"]
        self.proxiable = kwargs["proxiable"]
        self.address_restricted = kwargs["address_restricted"]
        self.canonicalization = kwargs["canonicalization"]
        self.enterprise = kwargs["enterprise"]
        self.renewal = kwargs["renewal"]
        self.validate = kwargs["validate"]
        self.keytab = kwargs["keytab"]
        self.keytab_path = kwargs["keytab_path"]

        self.kinit = CmdRunner(
            module,
            command="kinit",
            arg_formats=dict(
                lifetime=cmd_runner_fmt.as_opt_val("-l"),
                start_time=cmd_runner_fmt.as_opt_val("-s"),
                renewable=cmd_runner_fmt.as_opt_val("-r"),
                forwardable=cmd_runner_fmt.as_bool("-f", "-F", ignore_none=True),
                proxiable=cmd_runner_fmt.as_bool("-p", "-P", ignore_none=True),
                address_restricted=cmd_runner_fmt.as_bool("-a", "-A", ignore_none=True),
                anonymous=cmd_runner_fmt.as_bool("-n"),
                canonicalization=cmd_runner_fmt.as_bool("-C"),
                enterprise=cmd_runner_fmt.as_bool("-E"),
                renewal=cmd_runner_fmt.as_bool("-R"),
                validate=cmd_runner_fmt.as_bool("-v"),
                keytab=cmd_runner_fmt.as_bool("-k"),
                keytab_path=cmd_runner_fmt.as_func(lambda v: ["-t", v] if v else ["-i"]),
                cache_name=cmd_runner_fmt.as_opt_val("-c"),
                principal=cmd_runner_fmt.as_list(),
            ),
        )

        self.kdestroy = CmdRunner(
            module,
            command="kdestroy",
            arg_formats=dict(
                kdestroy_all=cmd_runner_fmt.as_bool("-A"),
                cache_name=cmd_runner_fmt.as_opt_val("-c"),
                principal=cmd_runner_fmt.as_opt_val("-p"),
            ),
        )

        self.klist = CmdRunner(
            module,
            command="klist",
            arg_formats=dict(
                show_list=cmd_runner_fmt.as_bool("-l"),
            ),
        )

    def exec_kinit(self):
        params = dict(self.module.params)
        with self.kinit(
            "lifetime start_time renewable forwardable proxiable address_restricted anonymous "
            "canonicalization enterprise renewal validate keytab keytab_path cache_name principal",
            check_rc=True,
            data=self.password,
        ) as ctx:
            rc, out, err = ctx.run(**params)
        return out

    def exec_kdestroy(self):
        params = dict(self.module.params)
        with self.kdestroy("kdestroy_all cache_name principal", check_rc=True) as ctx:
            rc, out, err = ctx.run(**params)
        return out

    def exec_klist(self, show_list):
        # Use chech_rc = False because
        # If no tickets present, klist command will always return rc = 1
        params = dict(show_list=show_list)
        with self.klist("show_list", check_rc=False) as ctx:
            rc, out, err = ctx.run(**params)
        return rc, out, err

    def check_ticket_present(self):
        ticket_present = True
        show_list = False

        if not self.principal and not self.cache_name:
            rc, out, err = self.exec_klist(show_list)
            if rc != 0:
                ticket_present = False
        else:
            show_list = True
            rc, out, err = self.exec_klist(show_list)
            if self.principal and self.principal not in str(out):
                ticket_present = False
            if self.cache_name and self.cache_name not in str(out):
                ticket_present = False

        return ticket_present


def main():
    arg_spec = dict(
        principal=dict(type="str"),
        password=dict(type="str", no_log=True),
        state=dict(default="present", choices=["present", "absent"]),
        kdestroy_all=dict(type="bool"),
        cache_name=dict(type="str", fallback=(env_fallback, ["KRB5CCNAME"])),
        lifetime=dict(type="str"),
        start_time=dict(type="str"),
        renewable=dict(type="str"),
        forwardable=dict(type="bool"),
        proxiable=dict(type="bool"),
        address_restricted=dict(type="bool"),
        anonymous=dict(type="bool"),
        canonicalization=dict(type="bool"),
        enterprise=dict(type="bool"),
        renewal=dict(type="bool"),
        validate=dict(type="bool"),
        keytab=dict(type="bool"),
        keytab_path=dict(type="path"),
    )
    module = AnsibleModule(
        argument_spec=arg_spec,
        supports_check_mode=True,
        required_by={"keytab_path": "keytab"},
        required_if=[
            ("state", "present", ("password", "keytab_path"), True),
        ],
    )

    state = module.params["state"]
    kdestroy_all = module.params["kdestroy_all"]

    keytab = IPAKeytab(
        module,
        state=state,
        kdestroy_all=kdestroy_all,
        principal=module.params["principal"],
        password=module.params["password"],
        cache_name=module.params["cache_name"],
        lifetime=module.params["lifetime"],
        start_time=module.params["start_time"],
        renewable=module.params["renewable"],
        forwardable=module.params["forwardable"],
        proxiable=module.params["proxiable"],
        address_restricted=module.params["address_restricted"],
        anonymous=module.params["anonymous"],
        canonicalization=module.params["canonicalization"],
        enterprise=module.params["enterprise"],
        renewal=module.params["renewal"],
        validate=module.params["validate"],
        keytab=module.params["keytab"],
        keytab_path=module.params["keytab_path"],
    )

    if module.params["keytab_path"] is not None and module.params["keytab"] is not True:
        module.fail_json(msg="If keytab_path is specified then keytab parameter must be True")

    changed = False
    if state == "present":
        if not keytab.check_ticket_present():
            changed = True
            if not module.check_mode:
                keytab.exec_kinit()

    if state == "absent":
        if kdestroy_all:
            changed = True
            if not module.check_mode:
                keytab.exec_kdestroy()
        elif keytab.check_ticket_present():
            changed = True
            if not module.check_mode:
                keytab.exec_kdestroy()

    module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_config

short_description: Manage docker configs

description:
  - Create and remove Docker configs in a Swarm environment. Similar to C(docker config create) and C(docker config rm).
  - Adds to the metadata of new configs C(ansible_key), an encrypted hash representation of the data, which is then used in
    future runs to test if a config has changed. If C(ansible_key) is not present, then a config will not be updated unless
    the O(force) option is set.
  - Updates to configs are performed by removing the config and creating it again.
extends_documentation_fragment:
  - community.docker._docker
  - community.docker._docker.docker_py_2_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  idempotent:
    support: partial
    details:
      - If O(force=true) the module is not idempotent.

options:
  data:
    description:
      - The value of the config.
      - Mutually exclusive with O(data_src). One of O(data) and O(data_src) is required if O(state=present).
    type: str
  data_is_b64:
    description:
      - If set to V(true), the data is assumed to be Base64 encoded and will be decoded before being used.
      - To use binary O(data), it is better to keep it Base64 encoded and let it be decoded by this option.
    type: bool
    default: false
  data_src:
    description:
      - The file on the target from which to read the config.
      - Mutually exclusive with O(data). One of O(data) and O(data_src) is required if O(state=present).
    type: path
    version_added: 1.10.0
  labels:
    description:
      - A map of key:value meta data, where both the C(key) and C(value) are expected to be a string.
      - If new meta data is provided, or existing meta data is modified, the config will be updated by removing it and creating
        it again.
    type: dict
  force:
    description:
      - Use with O(state=present) to always remove and recreate an existing config.
      - If V(true), an existing config will be replaced, even if it has not been changed.
    type: bool
    default: false
  rolling_versions:
    description:
      - If set to V(true), configs are created with an increasing version number appended to their name.
      - Adds a label containing the version number to the managed configs with the name C(ansible_version).
    type: bool
    default: false
    version_added: 2.2.0
  versions_to_keep:
    description:
      - When using O(rolling_versions), the number of old versions of the config to keep.
      - Extraneous old configs are deleted after the new one is created.
      - Set to V(-1) to keep everything or V(0) or V(1) to keep only the current one.
    type: int
    default: 5
    version_added: 2.2.0
  name:
    description:
      - The name of the config.
    type: str
    required: true
  state:
    description:
      - Set to V(present), if the config should exist, and V(absent), if it should not.
    type: str
    default: present
    choices:
      - absent
      - present
  template_driver:
    description:
      - Set to V(golang) to use a Go template in O(data) or a Go template file in O(data_src).
    type: str
    choices:
      - golang
    version_added: 2.5.0

requirements:
  - "L(Docker SDK for Python,https://docker-py.readthedocs.io/en/stable/) >= 2.6.0"
  - "Docker API >= 1.30"

author:
  - Chris Houseknecht (@chouseknecht)
  - John Hu (@ushuz)
"""

EXAMPLES = r"""
---
- name: Create config foo (from a file on the control machine)
  community.docker.docker_config:
    name: foo
    # If the file is JSON or binary, Ansible might modify it (because
    # it is first decoded and later re-encoded). Base64-encoding the
    # file directly after reading it prevents this to happen.
    data: "{{ lookup('file', '/path/to/config/file') | b64encode }}"
    data_is_b64: true
    state: present

- name: Create config foo (from a file on the target machine)
  community.docker.docker_config:
    name: foo
    data_src: /path/to/config/file
    state: present

- name: Change the config data
  community.docker.docker_config:
    name: foo
    data: Goodnight everyone!
    labels:
      bar: baz
      one: '1'
    state: present

- name: Add a new label
  community.docker.docker_config:
    name: foo
    data: Goodnight everyone!
    labels:
      bar: baz
      one: '1'
      # Adding a new label will cause a remove/create of the config
      two: '2'
    state: present

- name: No change
  community.docker.docker_config:
    name: foo
    data: Goodnight everyone!
    labels:
      bar: baz
      one: '1'
      # Even though 'two' is missing, there is no change to the existing config
    state: present

- name: Update an existing label
  community.docker.docker_config:
    name: foo
    data: Goodnight everyone!
    labels:
      bar: monkey # Changing a label will cause a remove/create of the config
      one: '1'
    state: present

- name: Force the (re-)creation of the config
  community.docker.docker_config:
    name: foo
    data: Goodnight everyone!
    force: true
    state: present

- name: Remove config foo
  community.docker.docker_config:
    name: foo
    state: absent
"""

RETURN = r"""
config_id:
  description:
    - The ID assigned by Docker to the config object.
  returned: success and O(state=present)
  type: str
  sample: 'hzehrmyjigmcp2gb6nlhmjqcv'
config_name:
  description:
    - The name of the created config object.
  returned: success and O(state=present)
  type: str
  sample: 'awesome_config'
  version_added: 2.2.0
"""

import base64
import hashlib
import traceback
import typing as t

try:
    from docker.errors import APIError, DockerException
except ImportError:
    # missing Docker SDK for Python handled in ansible.module_utils.docker.common
    pass

from ansible.module_utils.common.text.converters import to_bytes

from ansible_collections.community.docker.plugins.module_utils._common import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DockerBaseClass,
    compare_generic,
    sanitize_labels,
)


class ConfigManager(DockerBaseClass):
    def __init__(self, client: AnsibleDockerClient, results: dict[str, t.Any]) -> None:
        super().__init__()

        self.client = client
        self.results = results
        self.check_mode = self.client.check_mode

        parameters = self.client.module.params
        self.name = parameters.get("name")
        self.state = parameters.get("state")
        self.data = parameters.get("data")
        if self.data is not None:
            if parameters.get("data_is_b64"):
                self.data = base64.b64decode(self.data)
            else:
                self.data = to_bytes(self.data)
        data_src = parameters.get("data_src")
        if data_src is not None:
            try:
                with open(data_src, "rb") as f:
                    self.data = f.read()
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.client.fail(f"Error while reading {data_src}: {exc}")
        self.labels = parameters.get("labels")
        self.force = parameters.get("force")
        self.rolling_versions = parameters.get("rolling_versions")
        self.versions_to_keep = parameters.get("versions_to_keep")
        self.template_driver = parameters.get("template_driver")

        if self.rolling_versions:
            self.version = 0
        self.data_key: str | None = None
        self.configs: list[dict[str, t.Any]] = []

    def __call__(self) -> None:
        self.get_config()
        if self.state == "present":
            self.data_key = hashlib.sha224(self.data).hexdigest()
            self.present()
            self.remove_old_versions()
        elif self.state == "absent":
            self.absent()

    def get_version(self, config: dict[str, t.Any]) -> int:
        try:
            return int(
                config.get("Spec", {}).get("Labels", {}).get("ansible_version", 0)
            )
        except ValueError:
            return 0

    def remove_old_versions(self) -> None:
        if not self.rolling_versions or self.versions_to_keep < 0:
            return
        if not self.check_mode:
            while len(self.configs) > max(self.versions_to_keep, 1):
                self.remove_config(self.configs.pop(0))

    def get_config(self) -> None:
        """Find an existing config."""
        try:
            configs = self.client.configs(filters={"name": self.name})
        except APIError as exc:
            self.client.fail(f"Error accessing config {self.name}: {exc}")

        if self.rolling_versions:
            self.configs = [
                config
                for config in configs
                if config["Spec"]["Name"].startswith(f"{self.name}_v")
            ]
            self.configs.sort(key=self.get_version)
        else:
            self.configs = [
                config for config in configs if config["Spec"]["Name"] == self.name
            ]

    def create_config(self) -> str | None:
        """Create a new config"""
        config_id: str | dict[str, t.Any] | None = None
        # We ca not see the data after creation, so adding a label we can use for idempotency check
        labels = {"ansible_key": self.data_key}
        if self.rolling_versions:
            self.version += 1
            labels["ansible_version"] = str(self.version)
            self.name = f"{self.name}_v{self.version}"
        if self.labels:
            labels.update(self.labels)

        try:
            if not self.check_mode:
                # only use templating argument when self.template_driver is defined
                kwargs = {}
                if self.template_driver:
                    kwargs["templating"] = {"name": self.template_driver}
                config_id = self.client.create_config(
                    self.name, self.data, labels=labels, **kwargs
                )
                self.configs += self.client.configs(filters={"id": config_id})
        except APIError as exc:
            self.client.fail(f"Error creating config: {exc}")

        if isinstance(config_id, dict):
            return config_id["ID"]

        return config_id

    def remove_config(self, config: dict[str, t.Any]) -> None:
        try:
            if not self.check_mode:
                self.client.remove_config(config["ID"])
        except APIError as exc:
            self.client.fail(f"Error removing config {config['Spec']['Name']}: {exc}")

    def present(self) -> None:
        """Handles state == 'present', creating or updating the config"""
        if self.configs:
            config = self.configs[-1]
            self.results["config_id"] = config["ID"]
            self.results["config_name"] = config["Spec"]["Name"]
            data_changed = False
            template_driver_changed = False
            attrs = config.get("Spec", {})
            if attrs.get("Labels", {}).get("ansible_key"):
                if attrs["Labels"]["ansible_key"] != self.data_key:
                    data_changed = True
            else:
                if not self.force:
                    self.client.module.warn(
                        "'ansible_key' label not found. Config will not be changed unless the force parameter is set to 'true'"
                    )
            # template_driver has changed if it was set in the previous config
            # and now it differs, or if it was not set but now it is.
            if attrs.get("Templating", {}).get("Name"):
                if attrs["Templating"]["Name"] != self.template_driver:
                    template_driver_changed = True
            elif self.template_driver:
                template_driver_changed = True
            labels_changed = not compare_generic(
                self.labels, attrs.get("Labels"), "allow_more_present", "dict"
            )
            if self.rolling_versions:
                self.version = self.get_version(config)
            if data_changed or template_driver_changed or labels_changed or self.force:
                # if something changed or force, delete and re-create the config
                if not self.rolling_versions:
                    self.absent()
                config_id = self.create_config()
                self.results["changed"] = True
                self.results["config_id"] = config_id
                self.results["config_name"] = self.name
        else:
            self.results["changed"] = True
            self.results["config_id"] = self.create_config()
            self.results["config_name"] = self.name

    def absent(self) -> None:
        """Handles state == 'absent', removing the config"""
        if self.configs:
            for config in self.configs:
                self.remove_config(config)
            self.results["changed"] = True


def main() -> None:
    argument_spec = {
        "name": {"type": "str", "required": True},
        "state": {
            "type": "str",
            "default": "present",
            "choices": ["absent", "present"],
        },
        "data": {"type": "str"},
        "data_is_b64": {"type": "bool", "default": False},
        "data_src": {"type": "path"},
        "labels": {"type": "dict"},
        "force": {"type": "bool", "default": False},
        "rolling_versions": {"type": "bool", "default": False},
        "versions_to_keep": {"type": "int", "default": 5},
        "template_driver": {"type": "str", "choices": ["golang"]},
    }

    required_if = [
        ("state", "present", ["data", "data_src"], True),
    ]

    mutually_exclusive = [
        ("data", "data_src"),
    ]

    option_minimal_versions = {
        "template_driver": {"docker_py_version": "5.0.3", "docker_api_version": "1.37"},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=required_if,
        mutually_exclusive=mutually_exclusive,
        min_docker_version="2.6.0",
        min_docker_api_version="1.30",
        option_minimal_versions=option_minimal_versions,
    )
    sanitize_labels(client.module.params["labels"], "labels", client)

    try:
        results = {
            "changed": False,
        }

        ConfigManager(client, results)()
        client.module.exit_json(**results)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when Docker SDK for Python tried to talk to the docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_postgresqlflexibleserver_info
version_added: "2.2.0"
short_description: Get Azure PostgreSQL Flexible Server facts
description:
    - Get facts of PostgreSQL Flexible Server.

options:
    resource_group:
        description:
            - The name of the resource group that contains the resource. You can obtain this value from the Azure Resource Manager API or the portal.
        type: str
    name:
        description:
            - The name of the server.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: Get instance of PostgreSQL Flexible Server
  azure_rm_postgresqlflexibleserver_info:
    resource_group: myResourceGroup
    name: server_name

- name: List instances of PostgreSQL Flexible Server
  azure_rm_postgresqlflexibleserver_info:
    resource_group: myResourceGroup
    tags:
      - key
'''

RETURN = '''
servers:
    description:
        - A list of dictionaries containing facts for PostgreSQL Flexible servers.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID of the postgresql flexible server.
            returned: always
            type: str
            sample: "/subscriptions/xxx/resourceGroups/myResourceGroup/providers/Microsoft.DBforPostgreSQL/flexibleservers/postgresql3"
        resource_group:
            description:
                - Resource group name.
            returned: always
            type: str
            sample: myResourceGroup
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: postgreabdud1223
        location:
            description:
                - The location the resource resides in.
            returned: always
            type: str
            sample: eastus
        sku:
            description:
                - The SKU of the server.
            returned: always
            type: complex
            contains:
                name:
                    description:
                        - The name of the SKU.
                    returned: always
                    type: str
                    sample: Standard_B1ms
                tier:
                    description:
                        - The tier of the particular SKU.
                    returned: always
                    type: str
                    sample: Burstable
        storage:
            description:
                - The maximum storage allowed for a server.
            returned: always
            type: complex
            contains:
                storage_size_gb:
                    description:
                        - Max storage allowed for a server.
                    type: int
                    returned: always
                    sample: 128
        administrator_login:
            description:
                - The administrator's login name of a server.
            returned: always
            type: str
            sample: azureuser
        version:
            description:
                - Flexible Server version.
            returned: always
            type: str
            sample: "12"
        fully_qualified_domain_name:
            description:
                - The fully qualified domain name of the flexible server.
            returned: always
            type: str
            sample: postflexiblefredpgsqlflexible.postgres.database.azure.com
        availability_zone:
            description:
                - availability zone information of the server.
            type: str
            returned: always
            sample: 1
        backup:
            description:
                - Backup properties of a server.
            type: complex
            returned: always
            contains:
                backup_retention_days:
                    description:
                        - Backup retention days for the server.
                    type: int
                    returned: always
                    sample: 7
                geo_redundant_backup:
                    description:
                        - A value indicating whether Geo-Redundant backup is enabled on the server.
                    type: str
                    returned: always
                    sample: Disabled
        high_availability:
            description:
                - High availability properties of a server.
            type: complex
            returned: always
            contains:
                mode:
                    description:
                        - The HA mode for the server.
                    returned: always
                    sample: Disabled
                    type: str
                standby_availability_zone:
                    description:
                        - availability zone information of the standby.
                    type: str
                    returned: always
                    sample: null
        maintenance_window:
            description:
                - Maintenance window properties of a server.
            type: complex
            returned: always
            contains:
                custom_window:
                    description:
                        - Indicates whether custom window is enabled or disabled.
                    returned: always
                    sample: Enabled
                    type: str
                day_of_week:
                    description:
                        - Day of week for maintenance window.
                    returned: always
                    sample: 0
                    type: int
                start_hour:
                    description:
                        - Start hour for maintenance window.
                    type: int
                    returned: always
                    sample: 8
                start_minute:
                    description:
                        - Start minute for maintenance window.
                    type: int
                    returned: always
                    sample: 0
        network:
            description:
                - Network properties of a server.
            type: complex
            returned: always
            contains:
                delegated_subnet_resource_id:
                    description:
                        - Delegated subnet arm resource id.
                    type: str
                    returned: always
                    sample: null
                private_dns_zone_arm_resource_id:
                    description:
                        - Private dns zone arm resource id.
                    type: str
                    returned: always
                    sample: null
                public_network_access:
                    description:
                        - Public network access is enabled or not.
                    type: str
                    returned: always
                    sample: Enabled
        point_in_time_utc:
            description:
                - Restore point creation time (ISO8601 format).
            type: str
            sample: null
            returned: always
        source_server_resource_id:
            description:
                - The source server resource ID to restore from.
            type: str
            returned: always
            sample: null
        system_data:
            description:
                - The system metadata relating to this resource.
            type: complex
            returned: always
            contains:
                created_by:
                    description:
                        - The identity that created the resource.
                    type: str
                    returned: always
                    sample: null
                created_by_type:
                    description:
                        - The type of identity that created the resource.
                    returned: always
                    type: str
                    sample: null
                created_at:
                    description:
                        - The timestamp of resource creation (UTC).
                    returned: always
                    sample: null
                    type: str
                last_modified_by:
                    description:
                        - The identity that last modified the resource.
                    type: str
                    returned: always
                    sample: null
                last_modified_by_type:
                    description:
                        - The type of identity that last modified the resource.
                    returned: always
                    sample: null
                    type: str
                last_modified_at:
                    description:
                        - The timestamp of resource last modification (UTC).
                    returned: always
                    sample: null
                    type: str
        identity:
            description:
                - Identity for the Server.
            type: complex
            returned: when available
            contains:
                type:
                    description:
                        - Type of the managed identity
                    returned: always
                    sample: UserAssigned
                    type: str
                user_assigned_identities:
                    description:
                        - User Assigned Managed Identities and its options
                    returned: always
                    type: complex
                    contains:
                        id:
                            description:
                                - Dict of the user assigned identities IDs associated to the Resource
                            returned: always
                            type: dict
                            elements: dict
        tags:
            description:
                - Tags assigned to the resource. Dictionary of string:string pairs.
            type: dict
            returned: always
            sample: { tag1: abc }
        auth_config:
            description:
                - AuthConfig properties of a server.
            type: dict
            returned: always
            sample: {"active_directory_auth": "Disabled", "password_auth": "Enabled", "tenant_id": null}
'''


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
    import azure.mgmt.rdbms.postgresql_flexibleservers.models as PostgreSQLFlexibleModels
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMPostgreSqlFlexibleServersInfo(AzureRMModuleBase):
    def __init__(self):
        # define user inputs into argument
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
            ),
            name=dict(
                type='str'
            ),
            tags=dict(
                type='list',
                elements='str'
            )
        )
        # store the results of the module operation
        self.results = dict(
            changed=False
        )
        self.resource_group = None
        self.name = None
        self.tags = None
        super(AzureRMPostgreSqlFlexibleServersInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.resource_group is not None and self.name is not None:
            self.results['servers'] = self.get()
        elif self.resource_group is not None:
            self.results['servers'] = self.list_by_resource_group()
        else:
            self.results['servers'] = self.list_all()
        return self.results

    def get(self):
        response = None
        results = []
        try:
            response = self.postgresql_flexible_client.servers.get(resource_group_name=self.resource_group,
                                                                   server_name=self.name)
            self.log("Response : {0}".format(response))
        except ResourceNotFoundError:
            self.log('Could not get facts for PostgreSQL Flexible Server.')

        if response and self.has_tags(response.tags, self.tags):
            results.append(self.format_item(response))

        return results

    def list_by_resource_group(self):
        response = None
        results = []
        try:
            response = self.postgresql_flexible_client.servers.list_by_resource_group(resource_group_name=self.resource_group)
            self.log("Response : {0}".format(response))
        except Exception:
            self.log('Could not get facts for PostgreSQL Flexible Servers.')

        if response is not None:
            for item in response:
                if self.has_tags(item.tags, self.tags):
                    results.append(self.format_item(item))

        return results

    def list_all(self):
        response = None
        results = []
        try:
            response = self.postgresql_flexible_client.servers.list()
            self.log("Response : {0}".format(response))
        except Exception:
            self.log('Could not get facts for PostgreSQL Flexible Servers.')

        if response is not None:
            for item in response:
                if self.has_tags(item.tags, self.tags):
                    results.append(self.format_item(item))

        return results

    def format_item(self, item):
        result = dict(
            id=item.id,
            resource_group=self.resource_group,
            name=item.name,
            sku=dict(),
            location=item.location,
            tags=item.tags,
            system_data=dict(),
            administrator_login=item.administrator_login,
            version=item.version,
            minor_version=item.minor_version,
            fully_qualified_domain_name=item.fully_qualified_domain_name,
            storage=dict(),
            backup=dict(),
            network=dict(),
            high_availability=dict(),
            maintenance_window=dict(),
            source_server_resource_id=item.source_server_resource_id,
            point_in_time_utc=item.point_in_time_utc,
            availability_zone=item.availability_zone,
            auth_config=dict()
        )
        if item.sku is not None:
            result['sku']['name'] = item.sku.name
            result['sku']['tier'] = item.sku.tier
        if item.system_data is not None:
            result['system_data']['created_by'] = item.system_data.created_by
            result['system_data']['created_by_type'] = item.system_data.created_by_type
            result['system_data']['created_at'] = item.system_data.created_at
            result['system_data']['last_modified_by'] = item.system_data.last_modified_by
            result['system_data']['last_modified_by_type'] = item.system_data.last_modified_by_type
            result['system_data']['last_modified_at'] = item.system_data.last_modified_at
        if item.storage is not None:
            result['storage']['storage_size_gb'] = item.storage.storage_size_gb
        if item.backup is not None:
            result['backup']['backup_retention_days'] = item.backup.backup_retention_days
            result['backup']['geo_redundant_backup'] = item.backup.geo_redundant_backup
        if item.network is not None:
            result['network']['public_network_access'] = item.network.public_network_access
            result['network']['delegated_subnet_resource_id'] = item.network.delegated_subnet_resource_id
            result['network']['private_dns_zone_arm_resource_id'] = item.network.private_dns_zone_arm_resource_id
        if item.high_availability is not None:
            result['high_availability']['mode'] = item.high_availability.mode
            result['high_availability']['standby_availability_zone'] = item.high_availability.standby_availability_zone
        if item.maintenance_window is not None:
            result['maintenance_window']['custom_window'] = item.maintenance_window.custom_window
            result['maintenance_window']['start_minute'] = item.maintenance_window.start_minute
            result['maintenance_window']['start_hour'] = item.maintenance_window.start_hour
            result['maintenance_window']['day_of_week'] = item.maintenance_window.day_of_week
        if item.identity is not None:
            result['identity'] = item.identity.as_dict()
        else:
            result['identity'] = PostgreSQLFlexibleModels.UserAssignedIdentity(type='None').as_dict()
        if item.auth_config:
            result['auth_config']['active_directory_auth'] = item.auth_config.active_directory_auth
            result['auth_config']['password_auth'] = item.auth_config.password_auth
            result['auth_config']['tenant_id'] = item.auth_config.tenant_id

        return result


def main():
    AzureRMPostgreSqlFlexibleServersInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_realm_users_info

short_description: Retrieve users from a Keycloak realm using the Keycloak API

version_added: 13.1.0

description:
  - This module retrieves all users from a specified Keycloak realm using the Keycloak REST API.
  - Access to the REST API is performed via OpenID Connect. The user and client used must have the necessary permissions.
  - Authentication can be performed either with username/password or with a token.
  - The names of module options are snake_case versions of the camelCase ones found in the Keycloak API
    and its documentation at U(https://www.keycloak.org/docs-api/18.0/rest-api/index.html).

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  realm:
    type: str
    description:
      - The Keycloak realm from which users should be retrieved.
    default: 'master'

extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes
  - community.general._attributes.info_module

author:
  - Felix Grzelka (@felix-grzelka)
"""

EXAMPLES = r"""
- name: List all users in the "MyCustomRealm" realm using username/password authentication
  community.general.keycloak_realm_users_info:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
  delegate_to: localhost

- name: List all users in the "MyCustomRealm" realm using a token
  community.general.keycloak_realm_users_info:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
  delegate_to: localhost
"""

RETURN = r"""
users:
  description: List of users in the specified realm.
  returned: always
  type: list
  elements: dict
  sample:
    - id: "1234-5678-90"
      username: "user1"
      email: "user1@example.com"
    - id: "2345-6789-01"
      username: "user2"
      email: "user2@example.com"
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    argument_spec = keycloak_argument_spec()

    argument_spec["realm"] = dict(default="master")

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"]]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", users="")

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")

    result["users"] = kc.get_realm_users(realm=realm)
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: dynamodb_table_info
version_added: 7.2.0
short_description: Returns information about a Dynamo DB table
description:
  - Returns information about the Dynamo DB table, including the current status of the table,
    when it was created, the primary key schema, and any indexes on the table.
author:
  - Aubin Bikouo (@abikouo)
options:
  name:
    description:
      - The name of the table to describe.
    required: true
    type: str
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Return information about the DynamoDB table named 'my-table'
  community.aws.dynamodb_table_info:
    name: my-table
"""

RETURN = r"""
table:
    description: The returned table params from the describe API call.
    returned: success
    type: complex
    contains:
        table_name:
            description: The name of the table.
            returned: always
            type: str
        table_status:
            description: The current state of the table.
            returned: always
            type: str
            sample: 'ACTIVE'
        creation_date_time:
            description: The date and time when the table was created, in UNIX epoch time format.
            returned: always
            type: str
        table_size_bytes:
            description: The total size of the specified table, in bytes.
            returned: always
            type: int
        item_count:
            description: The number of items in the specified table.
            returned: always
            type: int
        table_arn:
            description: The Amazon Resource Name (ARN) that uniquely identifies the table.
            returned: always
            type: str
        table_id:
            description: Unique identifier for the table for which the backup was created.
            returned: always
            type: str
        attribute_definitions:
            description: A list of attributes for describing the key schema for the table and indexes.
            returned: always
            type: complex
            contains:
                attribute_name:
                    description: A name for the attribute.
                    type: str
                    returned: always
                attribute_type:
                    description: The data type for the attribute, S (String), N (Number) and B (Binary).
                    type: str
                    returned: always
        key_schema:
            description: A list of key schemas that specify the attributes that make up the primary key of a table, or the key attributes of an index.
            returned: always
            type: complex
            contains:
                attribute_name:
                    description: The name of a key attribute.
                    type: str
                    returned: always
                key_type:
                    description: The role that this key attribute will assume, 'HASH' for partition key, 'RANGE' for sort key
                    type: str
                    returned: always
        billing_mode:
            description: Controls how you are charged for read and write throughput and how you manage capacity.
            returned: always
            type: str
        local_secondary_indexes:
            description: Represents one or more local secondary indexes on the table.
            returned: if any, on the table
            type: list
            elements: dict
        global_secondary_indexes:
            description: The global secondary indexes of table.
            returned: if any, on the table
            type: list
            elements: dict
        stream_specification:
            description:  The current DynamoDB Streams configuration for the table.
            returned: if any, on the table
            type: complex
            contains:
                stream_enabled:
                    description: Indicates whether DynamoDB Streams is enabled (true) or disabled (false) on the table.
                    type: bool
                    returned: always
                    sample: true
                stream_view_type:
                    description: When an item in the table is modified, stream_view_type determines what information is written to the stream for this table.
                    type: str
                    returned: always
                    sample: KEYS_ONLY
        latest_stream_label:
            description: A timestamp, in ISO 8601 format, for this stream.
            type: str
            returned: if any on the table
        latest_stream_arn:
            description: The Amazon Resource Name (ARN) that uniquely identifies the latest stream for this table.
            returned: if any on the table
            type: str
        global_table_version:
            description: Represents the version of global tables in use, if the table is replicated across AWS Regions.
            type: str
            returned: if the table is replicated
        replicas:
            description: Represents replicas of the table.
            type: list
            elements: dict
            returned: if any on the table
        source_backup_arn:
            description: The Amazon Resource Name (ARN) of the backup from which the table was restored.
            type: str
            returned: if any, on the table
        source_table_arn:
            description: The ARN of the source table of the backup that is being restored.
            type: str
            returned: if any, on the table
        restore_date_time:
            description: Point in time or source backup time.
            type: str
            returned: if any, on table
        restore_in_progress:
            description: Indicates if a restore is in progress or not.
            type: bool
            returned: if any, on table
        sse_description:
            description: The description of the server-side encryption status on the specified table.
            type: dict
            returned: if any, on table
            sample: {}
        archival_summary:
            description: Contains information about the table archive.
            type: complex
            returned: if any, on table
            contains:
                archival_date_time:
                    description:  The date and time when table archival was initiated by DynamoDB, in UNIX epoch time format.
                    type: str
                    returned: always
                archival_reason:
                    description: The reason DynamoDB archived the table.
                    type: str
                    returned: always
                    sample: INACCESSIBLE_ENCRYPTION_CREDENTIALS
                archival_backup_arn:
                    description: The Amazon Resource Name (ARN) of the backup the table was archived to, when applicable in the archival reason.
                    type: str
                    returned: always
        table_class:
            description: The table class of the specified table.
            type: str
            returned: if any on the table
            sample: STANDARD_INFREQUENT_ACCESS
        deletion_protection_enabled:
            description: Indicates whether deletion protection is enabled (true) or disabled (false) on the table.
            type: bool
            returned: always
            sample: true
        provisioned_throughput:
            description: The provisioned throughput settings for the table.
            type: dict
            returned: always
            sample: '{"number_of_decreases_today": 0, "read_capacity_units": 1, "write_capacity_units": 1}'
        tags:
            description: A dict of tags associated with the DynamoDB table.
            returned: always
            type: dict
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


# ResourceNotFoundException is expected here if the table doesn't exist
@AWSRetry.jittered_backoff(catch_extra_error_codes=["LimitExceededException", "ResourceInUseException"])
def _describe_table(client, **params):
    return client.describe_table(**params)


def describe_dynamodb_table(module):
    table_name = module.params.get("name")
    retry_decorator = AWSRetry.jittered_backoff(
        catch_extra_error_codes=["LimitExceededException", "ResourceInUseException", "ResourceNotFoundException"],
    )
    client = module.client("dynamodb", retry_decorator=retry_decorator)
    try:
        table = _describe_table(client, TableName=table_name)
    except is_boto3_error_code("ResourceNotFoundException"):
        module.exit_json(table={})
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Failed to describe table")

    table = table["Table"]
    try:
        tags = client.list_tags_of_resource(aws_retry=True, ResourceArn=table["TableArn"])["Tags"]
    except is_boto3_error_code("AccessDeniedException"):
        module.warn("Permission denied when listing tags")
        tags = []
    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.BotoCoreError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Failed to list table tags")

    table = camel_dict_to_snake_dict(table)
    table["tags"] = boto3_tag_list_to_ansible_dict(tags)

    if "table_class_summary" in table:
        table["table_class"] = table["table_class_summary"]["table_class"]
        del table["table_class_summary"]

    # billing_mode_summary doesn't always seem to be set but is always set for PAY_PER_REQUEST
    # and when updating the billing_mode
    if "billing_mode_summary" in table:
        table["billing_mode"] = table["billing_mode_summary"]["billing_mode"]
        del table["billing_mode_summary"]
    else:
        table["billing_mode"] = "PROVISIONED"

    # Restore summary
    if "restore_summary" in table:
        table["source_backup_arn"] = table["restore_summary"].get("source_backup_arn", "")
        table["source_table_arn"] = table["restore_summary"].get("source_table_arn", "")
        table["restore_date_time"] = table["restore_summary"].get("restore_date_time", "")
        table["restore_in_progress"] = table["restore_summary"].get("restore_in_progress")
        del table["restore_summary"]

    module.exit_json(table=table)


def main():
    argument_spec = dict(
        name=dict(
            required=True,
        ),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    describe_dynamodb_table(module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


class ModuleDocFragment(object):
    DOCUMENTATION = r'''
options:
  login_host:
    description:
      - The same as the V(Client(host='...'\)) argument.
    type: str
    default: 'localhost'

  login_port:
    description:
      - The same as the V(Client(port='...'\)) argument.
      - If not passed, relies on the driver's default argument value.
    type: int

  login_db:
    description:
      - The same as the V(Client(database='...'\)) argument.
      - If not passed, relies on the driver's default argument value.
    type: str

  login_user:
    description:
      - The same as the V(Client(user='...'\)) argument.
      - If not passed, relies on the driver's default argument value.
      - Be sure your the user has permissions to read the system tables
        listed in the RETURN section.
    type: str

  login_password:
    description:
      - The same as the V(Client(password='...'\)) argument.
      - If not passed, relies on the driver's default argument value.
    type: str

  client_kwargs:
    description:
      - Any additional keyword arguments you want to pass
        to the Client interface when instantiating its object.
    type: dict
    default: {}

  success_on:
    version_added: "2.2.0"
    description:
      - List of L(server error codes,https://github.com/ClickHouse/ClickHouse/blob/master/src/Common/ErrorCodes.cpp)
        that will be treated as success, otherwise throw errors.
    type: list
    elements: int
    default: [497]

requirements: [ 'clickhouse-driver' ]

notes:
  - See the clickhouse-driver
    L(documentation,https://clickhouse-driver.readthedocs.io/en/latest)
    for more information about the driver interface.
'''



# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: openssl_privatekey_info
short_description: Retrieve information from OpenSSL private keys
version_added: 2.10.0
author:
  - Felix Fontein (@felixfontein)
description:
  - Provided an OpenSSL private keys, retrieve information.
  - This is a filter version of the M(community.crypto.openssl_privatekey_info) module.
options:
  _input:
    description:
      - The content of the OpenSSL private key.
    type: string
    required: true
  passphrase:
    description:
      - The passphrase for the private key.
    type: str
  return_private_key_data:
    description:
      - Whether to return private key data.
      - Only set this to V(true) when you want private information about this key to be extracted.
      - B(WARNING:) you have to make sure that private key data is not accidentally logged!
    type: bool
    default: false
extends_documentation_fragment:
  - community.crypto._name_encoding
seealso:
  - module: community.crypto.openssl_privatekey_info
"""

EXAMPLES = r"""
---
- name: Show the Subject Alt Names of the CSR
  ansible.builtin.debug:
    msg: >-
      {{
        (
          lookup('ansible.builtin.file', '/path/to/cert.csr')
          | community.crypto.openssl_privatekey_info
        ).subject_alt_name | join(', ')
      }}
"""

RETURN = r"""
_value:
  description:
    - Information on the certificate.
  type: dict
  contains:
    public_key:
      description: Private key's public key in PEM format.
      returned: success
      type: str
      sample: "-----BEGIN PUBLIC KEY-----\nMIICIjANBgkqhkiG9w0BAQEFAAOCAg8A..."
    public_key_fingerprints:
      description:
        - Fingerprints of private key's public key.
        - For every hash algorithm available, the fingerprint is computed.
      returned: success
      type: dict
      sample: "{'sha256': 'd4:b3:aa:6d:c8:04:ce:4e:ba:f6:29:4d:92:a3:94:b0:c2:ff:bd:bf:33:63:11:43:34:0f:51:b0:95:09:2f:63',
        'sha512': 'f7:07:4a:f0:b0:f0:e6:8b:95:5f:f9:e6:61:0a:32:68:f1..."
    type:
      description:
        - The key's type.
        - One of V(RSA), V(DSA), V(ECC), V(Ed25519), V(X25519), V(Ed448), or V(X448).
        - Will start with V(unknown) if the key type cannot be determined.
      returned: success
      type: str
      sample: RSA
    public_data:
      description:
        - Public key data. Depends on key type.
      returned: success
      type: dict
      contains:
        size:
          description:
            - Bit size of modulus (RSA) or prime number (DSA).
          type: int
          returned: When RV(_value.type=RSA) or RV(_value.type=DSA)
        modulus:
          description:
            - The RSA key's modulus.
          type: int
          returned: When RV(_value.type=RSA)
        exponent:
          description:
            - The RSA key's public exponent.
          type: int
          returned: When RV(_value.type=RSA)
        p:
          description:
            - The C(p) value for DSA.
            - This is the prime modulus upon which arithmetic takes place.
          type: int
          returned: When RV(_value.type=DSA)
        q:
          description:
            - The C(q) value for DSA.
            - This is a prime that divides C(p - 1), and at the same time the order of the subgroup of the multiplicative
              group of the prime field used.
          type: int
          returned: When RV(_value.type=DSA)
        g:
          description:
            - The C(g) value for DSA.
            - This is the element spanning the subgroup of the multiplicative group of the prime field used.
          type: int
          returned: When RV(_value.type=DSA)
        curve:
          description:
            - The curve's name for ECC.
          type: str
          returned: When RV(_value.type=ECC)
        exponent_size:
          description:
            - The maximum number of bits of a private key. This is basically the bit size of the subgroup used.
          type: int
          returned: When RV(_value.type=ECC)
        x:
          description:
            - The C(x) coordinate for the public point on the elliptic curve.
          type: int
          returned: When RV(_value.type=ECC)
        y:
          description:
            - For RV(_value.type=ECC), this is the C(y) coordinate for the public point on the elliptic curve.
            - For RV(_value.type=DSA), this is the publicly known group element whose discrete logarithm with respect to C(g)
              is the private key.
          type: int
          returned: When RV(_value.type=DSA) or RV(_value.type=ECC)
    private_data:
      description:
        - Private key data. Depends on key type.
      returned: success and when O(return_private_key_data) is set to V(true)
      type: dict
"""

import typing as t
from collections.abc import Callable

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.privatekey_info import (
    PrivateKeyParseError,
    get_privatekey_info,
)
from ansible_collections.community.crypto.plugins.plugin_utils._filter_module import (
    FilterModuleMock,
)


def openssl_privatekey_info_filter(
    data: str | bytes,
    passphrase: str | bytes | None = None,
    return_private_key_data: bool = False,
) -> dict[str, t.Any]:
    """Extract information from X.509 PEM certificate."""
    if not isinstance(data, (str, bytes)):
        raise AnsibleFilterError(
            f"The community.crypto.openssl_privatekey_info input must be a text type, not {type(data)}"
        )
    if passphrase is not None and not isinstance(passphrase, (str, bytes)):
        raise AnsibleFilterError(
            f"The passphrase option must be a text type, not {type(passphrase)}"
        )
    if not isinstance(return_private_key_data, bool):
        raise AnsibleFilterError(
            f"The return_private_key_data option must be a boolean, not {type(return_private_key_data)}"
        )

    module = FilterModuleMock({})
    try:
        result = get_privatekey_info(
            module=module,
            content=to_bytes(data),
            passphrase=to_text(passphrase) if passphrase is not None else None,
            return_private_key_data=return_private_key_data,
        )
        result.pop("can_parse_key", None)
        result.pop("key_is_consistent", None)
        return result
    except PrivateKeyParseError as exc:
        raise AnsibleFilterError(exc.error_message) from exc
    except OpenSSLObjectError as exc:
        raise AnsibleFilterError(str(exc)) from exc


class FilterModule:
    """Ansible jinja2 filters"""

    def filters(self) -> dict[str, Callable]:
        return {
            "openssl_privatekey_info": openssl_privatekey_info_filter,
        }
